/*
  The auth plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The auth plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The auth plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include "auth.h"
#include <osip/smsg.h>
#include <ppl/ppl_uinfo.h>

auth_ctx_t *auth_context = NULL;

extern psp_plugin_t PPL_DECLARE_DATA auth_plugin;

int
auth_ctx_init ()
{
  auth_context = (auth_ctx_t *) smalloc (sizeof (auth_ctx_t));
  if (auth_context == NULL)
    return -1;

  return 0;
}

void
auth_ctx_free ()
{
  if (auth_context == NULL)
    return;

  sfree (auth_context);
  auth_context = NULL;
}

int
auth_validate_credential_for_user (ppl_uinfo_t * user,
                                   proxy_authorization_t * p_auth, char *method)
{
  /* find the pending_auth element */
  char *response;
  char *nonce;
  char *opaque;
  char *realm;

  /* char *qop; */
  char *uri;

  nonce = proxy_authorization_getnonce (p_auth);
  opaque = proxy_authorization_getopaque (p_auth);
  if (opaque == NULL || nonce == NULL)
    return -1;

  realm = proxy_authorization_getrealm (p_auth);
  if (realm == NULL)
    return -1;
  response = proxy_authorization_getresponse (p_auth);
  if (response == NULL)
    return -1;
  uri = proxy_authorization_geturi (p_auth);
  if (uri == NULL)
    return -1;

  {
    /* verify credential */
    char *pszAlg = "MD5";
    char *pszUser = user->login;
    char *pszRealm = sgetcopy_unquoted_string (realm);
    char *pszPass = user->passwd;
    char *pszNonce = sgetcopy_unquoted_string (nonce);
    char *pszCNonce = NULL;

    char *szNonceCount = NULL;  /*  used with algo=md5-sess (not implemented) */
    char *pszMethod = method;
    char *pszURI = sgetcopy_unquoted_string (uri);
    char *pszQop = NULL;
    char *pszResponse = sgetcopy_unquoted_string (response);

    HASHHEX HA1;
    HASHHEX HA2 = "";
    HASHHEX Response;

    /*    if (qop!=NULL)
       pszQop = sgetcopy_unquoted_string(qop); */

    /* A1 = unq(username : unq(realm) : passwd ) */
    ppl_md5_DigestCalcHA1 (pszAlg, pszUser, pszRealm, pszPass, pszNonce,
                           pszCNonce, HA1);

    ppl_md5_DigestCalcResponse (HA1, pszNonce, szNonceCount, pszCNonce, pszQop,
                                pszMethod, pszURI, HA2, Response);

    sfree (pszRealm);
    sfree (pszNonce);
    sfree (pszURI);
    sfree (pszQop);

    OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                            "auth plugin: authentication response:!\nlocal:%s remote:%s\n",
                            pszResponse, Response));

    if (0 == strcmp (pszResponse, Response))
      {
        sfree (pszResponse);
        return 0;
      }
    sfree (pszResponse);
  }
  return -1;
}

ppl_uinfo_t *
auth_ctx_find_private_user (proxy_authorization_t * p_auth)
{
  ppl_uinfo_t *user;
  char *realm = psp_config_get_element ("serverrealm");
  char *username;
  char *qusername;

  if (0 != strcmp (p_auth->realm, realm))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                              "auth plugin: authentication header belongs to another proxy!\nlocal:%s remote:%s\n",
                              realm, p_auth->uri));
      return NULL;              /* this header is not for me */
    }

  /* search for a user context */
  qusername = proxy_authorization_getusername (p_auth);
  if (qusername == NULL)
    return NULL;
  username = sgetcopy_unquoted_string (qusername);
  user = ppl_uinfo_find_by_login (username);
  sfree (username);
  if (user != NULL)
    return user;
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                          "auth plugin: Could not find user: %s\n", username));
  return NULL;
}

/* HOOK METHODS */

/*
  This method returns:
  -2 if plugin consider this request should be totally discarded!
  -1 on error
  0  nothing has been done
  1  things has been done on psp_req element
*/
int
cb_auth_validate_credentials (psp_req_t * psp_req)
{
  ppl_uinfo_t *user;
  proxy_authorization_t *p_auth;
  authorization_t *h_auth;
  int i;
  int pos;

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                          "auth plugin: Processing authentication in request!\n"));

  if (!MSG_IS_REGISTER (psp_req->request))
    {
      /* default OUTPUT */
      psp_req_set_uas_status (psp_req, 407);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_state (psp_req, PSP_MANDATE);

      /* ANY proxy_authorization for us? */
      msg_getproxy_authorization (psp_req->request, 0, &p_auth);
      pos = 0;
      while (p_auth != NULL)    /* find the proxy_authorization for this proxy */
        {
          user = auth_ctx_find_private_user (p_auth);
          if (user != NULL)
            {
              i = auth_validate_credential_for_user (user, p_auth,
                                                     psp_req->request->strtline->
                                                     sipmethod);
              if (i != 0)
                {
                  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                                          "auth plugin: Bad credential for user!\n"));
                  /* may be we should answer 403 forbidden, because we have
                     detected that somebody tried to login! */
                  return 0;
              } else
                {
                  /* remove the authorization header! */
                  psp_req_set_state (psp_req, PSP_CONTINUE);
                  /* must remove the header */
                  list_remove (psp_req->request->proxy_authorizations, pos);
                  proxy_authorization_free (p_auth);
                  sfree (p_auth);
                }
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                                      "auth plugin: Valid credential for user!\n"));
              psp_req_set_mode (psp_req, PSP_SLESS_MODE);
              psp_req_set_state (psp_req, PSP_CONTINUE);
              return 0;         /* do nothing.. */
            }
          pos++;
          msg_getproxy_authorization (psp_req->request, pos, &p_auth);
        }
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                              "auth plugin: Bad credential for user!\n"));
      return 0;
    }

  /* for REGISTER */
  /* default OUTPUT */
  psp_req_set_uas_status (psp_req, 401);
  psp_req_set_mode (psp_req, PSP_UAS_MODE);
  psp_req_set_state (psp_req, PSP_MANDATE);

  /* ANY authorization for us? */
  h_auth = msg_getauthorization (psp_req->request);
  if (h_auth != NULL)           /* find the authorization for this proxy */
    {
      user = auth_ctx_find_private_user (h_auth);
      if (user != NULL)
        {
          i = auth_validate_credential_for_user (user,
                                                 (proxy_authorization_t *) h_auth,
                                                 psp_req->request->strtline->
                                                 sipmethod);
          if (i != 0)
            {                   /* invalid credentials! propose new one??? */
              OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                                      "auth plugin: Bad credential for user!\n"));
              return 0;
            }
          /* the rgstrar will answer 200 -if there is a registrar plugin- */
          psp_req_set_state (psp_req, PSP_CONTINUE);
          psp_req_set_mode (psp_req, PSP_SLESS_MODE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                                  "auth plugin: Valid credential for user!\n"));
          return 0;             /* do nothing.. */
        }
    }

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_WARNING, NULL,
                          "auth plugin: Bad credential for user!\n"));
  return 0;
}


int
cb_auth_add_credentials (psp_req_t * psp_req)   /* HOOK MIDDLE */
{
  char *nonce;
  char *opaque;

  /*  char *qop; */
  char *realm;
  proxy_authenticate_t *p_auth;
  www_authenticate_t *w_auth;
  int status = psp_req_get_uas_status (psp_req);

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                          "auth plugin: check if we need to add credentials in this 4xx!\n"));

  if (status != 407 && status != 401)
    {
      psp_req_set_state (psp_req, PSP_CONTINUE);
      return 0;
    }

  /* The request DOES not contains any credential */
  /* We have to challenge user */

  /* we just have to challenge the users */
  {
    MD5_CTX Md5Ctx;
    HASH HTMP;
    HASHHEX HTMPHex;
    int time_stamp;
    char *now;
    int i;

    /* build a nonce string */
    nonce = (char *) smalloc (HASHHEXLEN + 1 + 2);      /* +2 for the quotes */
    now = (char *) smalloc (30);
    time_stamp = ppl_time ();
    sprintf (now, "%i", time_stamp);
    ppl_MD5Init (&Md5Ctx);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) now, strlen (now));
    sfree (now);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);
    now = psp_config_get_element ("magicstring");
    ppl_MD5Update (&Md5Ctx, (unsigned char *) now, strlen (now));
    ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) psp_req->request->cseq->number,
                   strlen (psp_req->request->cseq->number));
    ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) psp_req->request->call_id->number,
                   strlen (psp_req->request->call_id->number));
    ppl_MD5Final ((unsigned char *) HTMP, &Md5Ctx);
    ppl_md5_hash_to_hex (HTMP, HTMPHex);
    sprintf (nonce, "\"%s\"", HTMPHex);

    /* build an opaque string */
    opaque = (char *) smalloc (HASHHEXLEN + 1 + 2);     /* +2 is for the quotes */
    now = (char *) smalloc (30);
    time_stamp = ppl_time ();
    sprintf (now, "%i", time_stamp);
    ppl_MD5Init (&Md5Ctx);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) now, strlen (now));
    sfree (now);
    ppl_MD5Update (&Md5Ctx, (unsigned char *) ":", 1);
    ppl_MD5Final ((unsigned char *) HTMP, &Md5Ctx);
    ppl_md5_hash_to_hex (HTMP, HTMPHex);
    sprintf (opaque, "\"%s\"", HTMPHex);

    /*    qop = smalloc(9);
       sprintf(qop, "\"auth\""); */

    realm = sgetcopy (psp_config_get_element ("serverrealm"));

    /* store data in a auth_ctx_t context */

    if (status == 407)
      {
        i = proxy_authenticate_init (&p_auth);
        if (i != 0)
          return -1;
        proxy_authenticate_setauth_type (p_auth, sgetcopy ("Digest"));
        /*      proxy_authenticate_setqop_options(p_auth, qop); */
        proxy_authenticate_setnonce (p_auth, nonce);
        proxy_authenticate_setopaque (p_auth, opaque);
        proxy_authenticate_setrealm (p_auth, realm);
        psp_req->response->proxy_authenticate = p_auth;

        psp_req_set_state (psp_req, PSP_CONTINUE);
        return 0;
    } else if (status == 401)
      {
        i = www_authenticate_init (&w_auth);
        if (i != 0)
          return -1;
        www_authenticate_setauth_type (w_auth, sgetcopy ("Digest"));
        /*      www_authenticate_setqop_options(w_auth, qop); */
        www_authenticate_setnonce (w_auth, nonce);
        www_authenticate_setopaque (w_auth, opaque);
        www_authenticate_setrealm (w_auth, realm);
        psp_req->response->www_authenticate = w_auth;

        psp_req_set_state (psp_req, PSP_CONTINUE);
        return 0;
      }
  }
  psp_req_set_state (psp_req, PSP_CONTINUE);
  return 0;
}
