/*
  The filter plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  the filter plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The filter plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include "filter.h"
#include <osip/smsg.h>
#include <ppl/ppl_uinfo.h>

filter_ctx_t *filter_context = NULL;

extern psp_plugin_t PPL_DECLARE_DATA filter_plugin;

int
filter_ctx_init ()
{
  filter_context = (filter_ctx_t *) smalloc (sizeof (filter_ctx_t));
  if (filter_context == NULL)
    return -1;

  return 0;
}

void
filter_ctx_free ()
{
  sfree (filter_context);
  filter_context = NULL;
}

/* HOOK METHODS */

/*
  This method returns:
  -2 if plugin consider this request should be totally discarded!
  -1 on error
  0  nothing has been done
  1  things has been done on psp_req element
*/
int
cb_filter_check_rules (psp_req_t * psp_req)
{

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                          "filter plugin: checking rules. (NOT IMPLEMENTED)\n"));
  /* default OUTPUTs for plugin */
  psp_req_set_mode (psp_req, PSP_SLESS_MODE);
  psp_req_set_state (psp_req, PSP_CONTINUE);

  /* TODO */


  return 0;
}

int
cb_filter_on_authentication (psp_req_t * psp_req)
{
  proxy_authorization_t *p_auth;
  authorization_t *h_auth;

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                          "filter plugin: start filtering request on authentication\n"));

  /* default OUTPUTs for plugin */
  psp_req_set_mode (psp_req, PSP_SLESS_MODE);
  psp_req_set_state (psp_req, PSP_CONTINUE);

  if (!MSG_IS_REGISTER (psp_req->request))
    {
      /* ANY proxy_authorization for us? */
      msg_getproxy_authorization (psp_req->request, 0, &p_auth);
      if (p_auth == NULL)       /* find the proxy_authorization for this proxy */
        {
          psp_req_set_uas_status (psp_req, 407);
          psp_req_set_mode (psp_req, PSP_UAS_MODE);
          psp_req_set_state (psp_req, PSP_MANDATE);
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                                  "filter plugin: No credential for user!\n"));
          return 0;
        }
      return 0;
    }

  /* for register request */
  h_auth = msg_getauthorization (psp_req->request);
  if (h_auth == NULL)           /* find the proxy_authorization for this proxy */
    {
      psp_req_set_uas_status (psp_req, 401);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_state (psp_req, PSP_MANDATE);
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "filter plugin: No credential for user!\n"));
      return 0;
    }
  return 0;
}
