/*
  The filter plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  the filter plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The filter plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>

#include "filter.h"

/* this structure is retreived by the core application
   with dlsym */
psp_plugin_t PSP_PLUGIN_DECLARE_DATA filter_plugin;

/* element shared with the core application layer */
/* all xxx_plugin elements MUST be DYNAMICLY ALLOCATED
   as the core application will assume they are */
imp_plugin_t *filter_plugin1;

extern filter_ctx_t *filter_context;

/* this is called by the core application when the module
   is loaded (the address of this method is located in
   the structure filter_plugin->plugin_init() */
PSP_PLUGIN_DECLARE (int)
plugin_init ()
{
  char *noauth;
  imp_func_t *fn;
  int i;

  /* plugin MUST create their own structure and give them
     back to the core by calling:
     psp_core_load_xxx_plugin();
     where xxx is the module name related to the plugin.
     psp_plugins can have more than one xxx_plugins attached
   */
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                          "filter plugin: plugin_init()!\n"));

  i = filter_ctx_init ();
  if (i != 0)
    goto pi_error1;

  psp_plugin_take_ownership (&filter_plugin);
  i = psp_core_load_imp_plugin (&filter_plugin1, &filter_plugin);
  if (i != 0)
    goto pi_error2;

  /* INIT HOOK METHOD FOR IMP */
  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_invite_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_register_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_ack_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_bye_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_options_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_info_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_cancel_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_notify_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_subscribe_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_check_rules, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_unknown_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  /* verify we actually want to do authentication */
  noauth = psp_config_get_element ("authentication");
  if (noauth == NULL)
    {
  } else if (0 == strcmp (noauth, "off"))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "filter plugin: Authentication is turned off!\n"));
      return 0;
    }
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                          "filter plugin: Authentication is turned on!\n"));

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_invite_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_register_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_ack_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_bye_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_options_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_info_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_cancel_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_notify_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_subscribe_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  i = imp_func_init (&fn, &cb_filter_on_authentication, filter_plugin.plug_id);
  if (i != 0)
    goto pi_error3;
  i = psp_core_add_imp_unknown_hook (fn, PSP_HOOK_REALLY_FIRST);
  if (i != 0)
    goto pi_error3;

  return 0;

pi_error3:
  /* ... */
pi_error2:
  filter_ctx_free ();
pi_error1:
  return -1;
}

PSP_PLUGIN_DECLARE (int)
plugin_start ()
{
  /*  OSIP_TRACE(osip_trace(__FILE__,__LINE__,OSIP_INFO1,NULL,
     "filter plugin: plugin_start()!\n"));
   */

  return -1;
}

PSP_PLUGIN_DECLARE (int)
plugin_release ()
{
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                          "filter plugin: plugin_release()!\n"));
  filter_ctx_free ();
  filter_context = NULL;
  return -1;
}

psp_plugin_t PSP_PLUGIN_DECLARE_DATA filter_plugin = {
  0,                            /* uninitialized */
  "Filter plugin",
  "0.4.5",
  "plugin with filtering capabilties.",
  PLUGIN_IMP,
  0,                            /* number of owners is always 0 at the begining */
  NULL,                         /* future place for the dso_handle */
  &plugin_init,
  &plugin_start,
  &plugin_release
};
