/*
  The ls_localdb plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The ls_localdb plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The ls_localdb plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include "ls_localdb.h"
#include <ppl/ppl_uinfo.h>

ls_localdb_ctx_t *ls_localdb_context = NULL;

extern psp_plugin_t PPL_DECLARE_DATA ls_localdb_plugin;


/*
  This plugin can be configured to:
  * act as a redirect server.
  *
  * It can also be configured to mandate record-routing.
  */

/*
  Configuration sample for REQUEST processing.

  <localdb>
  #  mode statefull
  #  mode redirect
  mode stateless
  record-route on
  </localdb>
*/

#define REDIRECT_MODE       0x01
#define STATELESS_MODE      0x02
#define STATEFULL_MODE      0x04


#define ISSET_REDIRECT_MODE(flag)   ((~flag|~REDIRECT_MODE)==~REDIRECT_MODE)
#define SET_REDIRECT_MODE(flag)     (flag=((flag&0x0FF)|REDIRECT_MODE))
#define ISSET_STATELESS_MODE(flag)  ((~flag|~STATELESS_MODE)==~STATELESS_MODE)
#define SET_STATELESS_MODE(flag)    (flag=((flag&0x0FF)|STATELESS_MODE))
#define ISSET_STATEFULL_MODE(flag)  ((~flag|~STATEFULL_MODE)==~STATEFULL_MODE)
#define SET_STATEFULL_MODE(flag)    (flag=((flag&0x0FF)|STATEFULL_MODE))

#define R_ROUTE_MODE        0x10

#define ISSET_R_ROUTE_MODE(flag)     ((~flag|~R_ROUTE_MODE)==~R_ROUTE_MODE)
#define SET_R_ROUTE_MODE(flag)       (flag= flag|R_ROUTE_MODE)

int
ls_localdb_ctx_init ()
{
  config_element_t *elem;

  ls_localdb_context = (ls_localdb_ctx_t *) smalloc (sizeof (ls_localdb_ctx_t));
  if (ls_localdb_context == NULL)
    return -1;

  ls_localdb_context->flag = 0;
  elem = psp_config_get_sub_element ("mode", "localdb", NULL);
  if (elem == NULL || elem->value == NULL)
    SET_STATEFULL_MODE (ls_localdb_context->flag);
  else if (0 == strcmp (elem->value, "redirect"))
    SET_REDIRECT_MODE (ls_localdb_context->flag);
  else if (0 == strcmp (elem->value, "stateless"))
    SET_STATELESS_MODE (ls_localdb_context->flag);
  else if (0 == strcmp (elem->value, "statefull"))
    SET_STATEFULL_MODE (ls_localdb_context->flag);
  else
    goto llci_error1;           /* error, bad option */

  elem = psp_config_get_sub_element ("record-route", "localdb", NULL);
  if (elem == NULL || elem->value == NULL)
    {                           /* default mode is off */
  } else if (0 == strcmp (elem->value, "off"))
    {
  } else if (0 == strcmp (elem->value, "on"))
    SET_R_ROUTE_MODE (ls_localdb_context->flag);
  else
    goto llci_error1;           /* error, bad option */

  if (ISSET_REDIRECT_MODE (ls_localdb_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_localdb plugin: configured to do redirect request!\n"));
    }
  if (ISSET_STATEFULL_MODE (ls_localdb_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_localdb plugin: configured to run in statefull mode!\n"));
    }
  if (ISSET_STATELESS_MODE (ls_localdb_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_localdb plugin: configured to run in stateless mode!\n"));
    }

  if (ISSET_R_ROUTE_MODE (ls_localdb_context->flag))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_localdb plugin: configured to do record-routing!\n"));
    }

  return 0;

llci_error1:
  ls_localdb_ctx_free ();
  return -1;
}

void
ls_localdb_ctx_free ()
{
  if (ls_localdb_context == NULL)
    return;

  sfree (ls_localdb_context);
  ls_localdb_context = NULL;
}

/* HOOK METHODS */

/*
  This method returns:
  -2 if plugin consider this request should be totally discarded!
  -1 on error
  0  nothing has been done
  1  things has been done on psp_req element
*/
int
cb_ls_localdb_search_user_location (psp_req_t * psp_req)
{
  route_t *route;
  ppl_uinfo_t *uinfo;
  int i;


  /* default OUTPUT */
  if (ISSET_R_ROUTE_MODE (ls_localdb_context->flag))
    psp_req_set_property (psp_req, PSP_STAY_ON_PATH);
  else
    psp_req_set_property (psp_req, 0);

  /* mode */
  if (ISSET_STATELESS_MODE (ls_localdb_context->flag))
    psp_req_set_mode (psp_req, PSP_SLESS_MODE);
  else if (ISSET_STATEFULL_MODE (ls_localdb_context->flag))
    psp_req_set_mode (psp_req, PSP_SFULL_MODE);
  else if (ISSET_REDIRECT_MODE (ls_localdb_context->flag))
    {
      psp_req_set_uas_status (psp_req, 302);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
  } else
    psp_req_set_mode (psp_req, PSP_SFULL_MODE);

  psp_req_set_state (psp_req, PSP_MANDATE);


  /* THIS IS MANDATORY FOR ALL PLUGINS TO CHECK THAT!
     this is rarely used as a previous plugin 'ls_localdb' should have
     checked for it. In case you remove it. */
  msg_getroute (psp_req->request, 1, &route);
  if (route != NULL)
    {
      psp_req_set_mode (psp_req, PSP_SLESS_MODE);
      psp_req_set_state (psp_req, PSP_MANDATE);
      return 0;                 /* do nothing.. */
    }

  if (psp_req->request->strtline->rquri->username == NULL)
    {
      /* another plugin may know how to answer this request?
         Let's keep searching... */
      psp_req_set_state (psp_req, PSP_CONTINUE);
      return 0;
    }

  /* look for this URI in our list of users */
  uinfo = ppl_uinfo_find_by_aor (psp_req->request->strtline->rquri);
  if (uinfo != NULL)
    {
      /* add all locations to the psp_req element */
      binding_t *bind;
      location_t *loc;
      url_t *url;

      bind = uinfo->bindings;
      for (; bind != NULL; bind = bind->next)
        {
          i = ppl_uinfo_check_binding (bind);
          if (i != 0)
            {                   /* binding is expired */
              ppl_uinfo_remove_binding (uinfo, bind);
            }
        }
      bind = uinfo->bindings;
      if (bind == NULL)
        {                       /* user is not around... */
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                  "ls_localdb plugin: User Exist but has no valid registration!\n"));
          psp_req_set_uas_status (psp_req, 480);
          psp_req_set_mode (psp_req, PSP_UAS_MODE);
          psp_req_set_state (psp_req, PSP_MANDATE);
          return 0;
        }
      i = url_clone (bind->contact->url, &url);
      if (i != 0)               /* bad configuration?->should be checked in plugin_init
                                   so this will never happen */
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                  "ls_localdb plugin: Could not clone contact info!\n"));
          psp_req_set_uas_status (psp_req, 400);
          psp_req_set_mode (psp_req, PSP_UAS_MODE);
          psp_req_set_state (psp_req, PSP_MANDATE);
          return -1;            /* error case (process can continue...) */
        }
      i = location_init (&loc, url, 3600);
      if (i != 0)               /* bad configuration?->should be checked in plugin_init
                                   so this will never happen */
        {
          /* This can only happen in case we don't have enough memory */
/* url_free(url); *//* dangerous? I prefer to leave a memory leak */
/* sfree(url); *//* as we may never enter this code */
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_BUG, NULL,
                                  "ls_localdb plugin: Could not create location info!\n"));
          psp_req_set_uas_status (psp_req, 400);
          psp_req_set_mode (psp_req, PSP_UAS_MODE);
          psp_req_set_state (psp_req, PSP_MANDATE);
          return -1;            /* error case (process can continue...) */
        }
      ADD_ELEMENT (psp_req->locations, loc);

      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                              "ls_localdb plugin: location found!\n"));
      /* TODO: When statefull mode will run, we'll change all return codes
         to ask for PSP_STATEFULL_MODE and will add more than one location
         in psp_req */
      return 0;
    }

  /* no user location found in local database... */
  psp_req_set_uas_status (psp_req, 404);
  psp_req_set_state (psp_req, PSP_CONTINUE);
  psp_req_set_mode (psp_req, PSP_UAS_MODE);
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                          "ls_localdb plugin: No location found for known user: return 404 Not found!!\n"));
  return 0;
}
