/*
  The syntax plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The syntax plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The syntax plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include "syntax.h"

extern psp_plugin_t PPL_DECLARE_DATA syntax_plugin;

/*
  0: default   <--- for INVITE: be statefull. For other request: be stateless
  1: no verification
  2: limited checks
  3: full checks */
int mode = 0;

/* HOOK METHODS */

/*
  This method returns:
  -2 if plugin consider this request should be totally discarded!
  -1 on error
  0  nothing has been done
  1  things has been done on psp_req element
*/
int
cb_check_syntax_in_request (psp_req_t * psp_req)
{
  header_t *maxfwd;
  header_t *p_require;

#ifdef WIN32
  unsigned long int one_inet_addr;
#else
  struct in_addr inaddr;
#endif

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO4, NULL,
                          "syntax plugin: validate syntax.\n"));

  /* verify validity of request just in case */
  if (psp_req == NULL
      || psp_req->request == NULL
      || psp_req->request->strtline == NULL
      || psp_req->request->strtline->rquri == NULL
      || (psp_req->request->strtline->rquri->host == NULL
          && psp_req->request->strtline->rquri->string == NULL))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: Bad Request-Line.\n"));
      psp_req_set_state (psp_req, PSP_STOP);
      return -2;                /* ask the core application to discard the request */
    }

  if (NULL == msg_getfrom (psp_req->request))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: No From header.\n"));
      psp_req_set_state (psp_req, PSP_STOP);
      return -2;
    }

  if (NULL == msg_getcall_id (psp_req->request))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: No Call-ID header.\n"));
      psp_req_set_state (psp_req, PSP_STOP);
      return -2;
    }

  if (NULL == msg_getcseq (psp_req->request))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: No CSeq header.\n"));
      psp_req_set_state (psp_req, PSP_STOP);
      return -2;
    }

  if (NULL == msg_getto (psp_req->request))
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: No To header.\n"));
      psp_req_set_state (psp_req, PSP_STOP);
      return -2;
    }

  if (NULL != psp_req->request->strtline->rquri->string)
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: Url Scheme not supported by proxy.\n"));
      /* by now, we only accept "sip:" and "sips:" urls */
      psp_req_set_state (psp_req, PSP_MANDATE);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_uas_status (psp_req, 416);
      return 0;
    }
#ifdef WIN32
  if ((int) (one_inet_addr == inet_addr (psp_req->request->strtline->rquri->host))
      == -1)
    {

    }
#else
  if (INADDR_NONE == inet_aton (psp_req->request->strtline->rquri->host, &inaddr))
    {
      /* TODO: check if it's an valid fqdn! */
      /*
         if (isavalidfqdn())
         {
         psp_req_set_state (psp_req, PSP_MANDATE);
         psp_req_set_mode (psp_req, PSP_UAS_MODE);
         psp_req_set_uas_status (psp_req, 400);
         }
         else
         {
         OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
         "syntax plugin: Ip address detected!"));
         }
       */
    }
#endif
  else
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                              "syntax plugin: Ip address detected!"));
    }
  msg_getmax_forward (psp_req->request, 0, &maxfwd);
  if (maxfwd != NULL)
    {
      if (maxfwd->hvalue != NULL
          && 1 == strlen (maxfwd->hvalue) && 0 == strncmp (maxfwd->hvalue, "0", 1))
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                  "syntax plugin: Too much Hop for request!"));
          /* by now, we only accept "sip:" and "sips:" urls */
          psp_req_set_state (psp_req, PSP_MANDATE);
          psp_req_set_mode (psp_req, PSP_UAS_MODE);
          psp_req_set_uas_status (psp_req, 483);        /* Too Many Hops */
          return 0;
        }
    }

  /* this is the correct place for doing loop detection and spiralling ! */


  /* Checking Proxy-Require extensions. */
  /* at this step, none is supported! */
  msg_getproxy_require (psp_req->request, 0, &p_require);
  if (p_require != NULL)
    {                           /* at least one exist, but we don't support any extensions */
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "syntax plugin: Unsupported feature in Proxy-Require header.\n"));
      psp_req_set_state (psp_req, PSP_MANDATE);
      psp_req_set_mode (psp_req, PSP_UAS_MODE);
      psp_req_set_uas_status (psp_req, 420);
      return 0;
    }

  psp_req_set_state (psp_req, PSP_CONTINUE);
  return 0;
}

int
cb_complete_answer_on_4xx (psp_req_t * psp_req)
{
  header_t *p_require;
  header_t *unsupported;
  int i;

  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO3, NULL,
                          "syntax plugin: completing 4xx answer.\n"));

  if (psp_req_get_uas_status (psp_req) == 420)  /* on Bad Extension */
    {
      int pos = 0;

      /* copy all supported header in unsupported headers */
      pos = msg_getproxy_require (psp_req->request, pos, &p_require);
      while (p_require != NULL)
        {
          if (p_require->hname == NULL || p_require->hvalue == NULL)
            break;
          i = header_clone (p_require, &unsupported);
          if (i != 0)
            break;
          sfree (unsupported->hname);
          unsupported->hname = sgetcopy ("Unsupported");
          list_add (psp_req->response->headers, unsupported, -1);
          pos++;
          pos = msg_getproxy_require (psp_req->request, pos, &p_require);
        }
    }
  return 0;
}
