/*
  The udp plugin is a GPL plugin for partysip.
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  The udp plugin is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  The udp plugin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>
#include <ppl/ppl_dns.h>
#include <ppl/ppl_socket.h>
#include "udp.h"

#include <fcntl.h>

local_ctx_t *ctx = NULL;

int
local_ctx_init (int in_port, int out_port)
{
  int i;
  struct sockaddr_in raddr;

  /*   int option=1; */

  ctx = (local_ctx_t *) smalloc (sizeof (local_ctx_t));
  if (ctx == NULL)
    return -1;

  ctx->in_port = in_port;

  /* not used by now */
  ctx->out_port = out_port;

  ctx->in_socket = ppl_socket (PF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (ctx->in_socket == -1)
    {
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "udp plugin: cannot create descriptor for port %i!\n",
                              ctx->in_port));
      goto lci_error1;
    }

  if (out_port == in_port)
    ctx->in_socket = ctx->out_socket;
  else
    {
      int try;

      ctx->out_socket = ppl_socket (PF_INET, SOCK_DGRAM, IPPROTO_UDP);
      if (ctx->out_socket == -1)
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                  "udp plugin: cannot create descriptor for port %i!\n",
                                  ctx->out_port));
          goto lci_error2;
        }
      i = (-1);
      try = 0;
      while (i < 0 && try < 40)
        {
          try++;
          raddr.sin_addr.s_addr = htons (INADDR_ANY);
          raddr.sin_port = htons ((short) ctx->out_port);
          raddr.sin_family = AF_INET;

          i = ppl_socket_bind (ctx->out_socket,
                               (struct sockaddr *) &raddr, sizeof (raddr));
          if (i < 0)
            {
#ifdef WIN32
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_WARNING, NULL,
                           "udp plugin: cannot bind on port %i!\n", ctx->in_port));
#else
              /* XXX use some configure magic to detect if we have the right strerror_r()
               * function (the one that returns int, not char *) and then use strerror_r()
               * instead.
               */
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_WARNING, NULL,
                           "udp plugin: cannot bind on port %i (%s)!\n",
                           ctx->in_port, strerror (errno)));
#endif
              ctx->out_port++;
            }
        }
      if (i != 0)
        goto lci_error3;
    }

  raddr.sin_addr.s_addr = htons (INADDR_ANY);
  raddr.sin_port = htons ((short) ctx->in_port);
  raddr.sin_family = AF_INET;

  if (ppl_socket_bind (ctx->in_socket,
                       (struct sockaddr *) &raddr, sizeof (raddr)) < 0)
    {
#ifdef WIN32
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "udp plugin: cannot bind on port %i!\n", ctx->in_port));
#else
      /* XXX use some configure magic to detect if we have the right strerror_r()
       * function (the one that returns int, not char *) and then use strerror_r()
       * instead.
       */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "udp plugin: cannot bind on port %i (%s)!\n", ctx->in_port,
                   strerror (errno)));
#endif
      goto lci_error4;
    }

  /* set the socket to never block on recv() calls */
#ifndef WIN32
  if (0 != fcntl (ctx->in_socket, F_SETFL, O_NONBLOCK))
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "udp plugin; cannot set O_NONBLOCK to the file desciptor (port=%i)!\n",
                   ctx->in_port));
      goto lci_error5;
    }
#else
  {
    int timeout = 0;
    int err;

    err = setsockopt (ctx->in_socket,
                      SOL_SOCKET,
                      SO_RCVTIMEO, (char *) &timeout, sizeof (timeout));
    if (err != NO_ERROR)
      {
        /* failed for some reason... */
        OSIP_TRACE (osip_trace
                    (__FILE__, __LINE__, OSIP_ERROR, NULL,
                     "udp plugin; cannot set O_NONBLOCK to the file desciptor (port=%i)!\n",
                     ctx->in_port));
        goto lci_error5;
      }
  }
#endif

  /*
     #ifdef __linux
     i = setsockopt (ctx->in_socket, SOL_SOCKET, SO_REUSEADDR,
     NULL, (socklen_t)0);
     #else
     i = setsockopt (ctx->in_socket, SOL_SOCKET, SO_REUSEADDR,
     NULL, 0);
     #endif
     if (i!=0)
     {
     OSIP_TRACE (osip_trace
     (__FILE__, __LINE__, OSIP_WARNING, NULL,
     "upd plugin; UDP listener SO_REUSE_ADDR failed %i (%i)!\n",
     ctx->in_port, i));
     }
   */
  return 0;

lci_error5:
lci_error4:
lci_error3:
  ppl_socket_close (ctx->out_socket);
lci_error2:
  ppl_socket_close (ctx->in_socket);
lci_error1:
  sfree (ctx);
  ctx = NULL;
  return -1;
}

void
local_ctx_free ()
{
  if (ctx == NULL)
    return;
  if (ctx->in_socket != -1)
    ppl_socket_close (ctx->in_socket);
  if (ctx->in_port == ctx->out_port)
    ctx->out_socket = (-1);
  else if (ctx->out_socket != -1)
    ppl_socket_close (ctx->out_socket);
  sfree (ctx);
  ctx = NULL;
}

/* max_analysed = maximum number of message when this method can parse
   messages whithout returning.

   This method returns:
   -1 on error
   0  on no message available
   1  on max_analysed reached
*/
int
cb_rcv_udp_message (int max)
{
  char *buf;
  int i;
  struct sockaddr_in sa;

#ifdef __linux
  socklen_t slen;
#else
  int slen;
#endif

  if (ctx == NULL)
    return -1;

  /* create a set of file descritor to control the stack */

  for (; max != 0; max--)
    {
      buf = (char *) smalloc (SIP_MESSAGE_MAX_LENGTH * sizeof (char) + 1);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "WAITING FOR UDP MESSAGE\n"));
      slen = sizeof (sa);
      i =
        ppl_socket_recv (ctx->in_socket, buf, SIP_MESSAGE_MAX_LENGTH, 0,
                         (struct sockaddr *) &sa, &slen);
      if (i > 0)
        {
          /* Message might not end with a "\0" but we know the number of */
          /* char received! */
          sstrncpy (buf + i, "\0", 1);
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_INFO2, NULL,
                       "udp plugin: RCV UDP MESSAGE (from %s:%i)\n",
                       inet_ntoa (sa.sin_addr), sa.sin_port));
#ifdef SHOW_MESSAGE
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_INFO1, NULL, "\n%s\n", buf));
#endif
          udp_process_message (buf);
      } else if (i == -1)
        {
          if (errno == EAGAIN)
            {
              sfree (buf);
              return 0;
            }
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "udp plugin: error while receiving data!\n"));
          sfree (buf);
          return -1;
        }
    }
  /* max is reached */
  return 1;
}

int
udp_process_message (char *buf)
{
  sipevent_t *evt;

  if (buf == NULL
      || *buf == '\0'
      || buf[1] == '\0'
      || buf[2] == '\0' || buf[3] == '\0' || buf[4] == '\0' || buf[5] == '\0')
    {
      sfree (buf);
      return -1;
    }

  evt = osip_parse (buf);
  sfree (buf);
  if (evt == NULL)
    /* discard... */
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "udp module: Could not parse response!\n"));
      return -1;
    }
  if (evt->sip == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "udp module: Could not parse response!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      return -1;
    }
  udp_log_event (evt);

  /* modify the request so it's compliant with the latest draft */
  psp_core_fix_strict_router_issue (evt);

  psp_core_event_add_sip_message (evt);
  return 0;
}


int
udp_log_event (sipevent_t * evt)
{
#ifdef SHOW_LIMITED_MESSAGE
  via_t *via;
  generic_param_t *b;

  via = list_get (evt->sip->vias, 0);
  via_param_getbyname (via, "branch", &b);
  if (b == NULL)
    {
      if (MSG_IS_REQUEST (evt->sip))
        {
      } else
        {
        }
      return -1;
    }
  if (MSG_IS_REQUEST (evt->sip))
    {
  } else
    {
    }
#endif
  return 0;
}

/* return
   -1 on error
   0  on success
*/
int
cb_snd_udp_message (const transaction_t * transaction,  /* read only element */
                    sip_t * message,    /* message to send           */
                    char *host, /* proposed destination host */
                    int port,   /* proposed destination port */
                    int socket) /* proposed socket (if any)  */
{
  int i;
  char *buf;
  struct hostent *hp;
  struct hostent **hp_ptr;
  struct sockaddr_in addr;
  unsigned long int one_inet_addr;
  int sock;

  if (ctx == NULL)
    return -1;

  hp = NULL;
  hp_ptr = NULL;

  i = msg_2char (message, &buf);

  if (i != 0)
    return -1;
#ifdef SHOW_MESSAGE
  OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL, "\n%s\n", buf));
#endif
  /* For RESPONSE, oSIP ALWAYS provide host and port from the top via */
  /* For REQUEST,  oSIP SOMETIMES provide host and port to use which
     may be different from the request uri */

  if (host == NULL)
    {                           /* when host is NULL, we use the request uri value */
      host = message->strtline->rquri->host;
      if (message->strtline->rquri->port != NULL)
        port = satoi (message->strtline->rquri->port);
      else
        port = 5060;
    }

  if ((int) (one_inet_addr = inet_addr (host)) == -1)
    {
      i = ppl_dns_gethostbyname (&addr, host, port);
      if (i != PPL_SUCCESS)
        {
          sfree (buf);
          return -1;
        }
  } else
    {
      addr.sin_port = htons ((short) port);
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = one_inet_addr;
    }

  sock = ctx->out_socket;

  /* connect(sock,(struct sockaddr *) &addr,sizeof(addr)); */

  if (0 > sendto (sock, (const void *) buf, strlen (buf), 0,
                  (struct sockaddr *) &addr, sizeof (addr)))
    {
      sfree (buf);
#ifdef WIN32
      i = WSAGetLastError ();
      if (WSAECONNREFUSED == i)
#else
      i = errno;
      if (ECONNREFUSED == i)
#endif
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                  "SIP_ECONNREFUSED - No remote server.\n"));
          return 1;             /* I prefer to answer 1 by now..
                                   we'll see later what's better */
        }
#ifdef WIN32
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "SIP_NETWORK_ERROR - Network error %i sending message to %s on port %i.\n",
                              i, host, port));
#else
      /* XXX use some configure magic to detect if we have the right strerror_r()
       * function (the one that returns int, not char *) and then use strerror_r()
       * instead.
       */
      OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_ERROR, NULL,
                              "SIP_NETWORK_ERROR - Network error %i (%s) sending message to %s on port %i.\n",
                              i, strerror (i), host, port));
#endif
      /* SIP_NETWORK_ERROR; */
      return -1;
    }

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "udp_plugin: message sent to %s on port %i\n", host, port));

  sfree (buf);
  return 0;
}
