/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>
#include <osip_ctx.h>
#include <osip_msg.h>

int
pspm_imp_init (pspm_imp_t ** pspm)
{
  int i;

  (*pspm) = (pspm_imp_t *) smalloc (sizeof (pspm_imp_t));
  if (*pspm == NULL)
    return -1;

  i = module_init (&((*pspm)->module), MOD_THREAD, "imp");
  if (i != 0)
    goto pii_error1;
  (*pspm)->imp_plugins = NULL;

  (*pspm)->sip_traffic = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sip_traffic == NULL)
    goto pii_error2;
  fifo_init ((*pspm)->sip_traffic);

  (*pspm)->func_invites = NULL;
  (*pspm)->func_acks = NULL;
  (*pspm)->func_registers = NULL;
  (*pspm)->func_byes = NULL;
  (*pspm)->func_optionss = NULL;
  (*pspm)->func_infos = NULL;
  (*pspm)->func_cancels = NULL;
  (*pspm)->func_notifys = NULL;
  (*pspm)->func_subscribes = NULL;
  (*pspm)->func_unknowns = NULL;

  i = imp_func_tab_init (&((*pspm)->func_invites));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_acks));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_registers));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_byes));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_optionss));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_infos));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_cancels));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_notifys));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_subscribes));
  if (i != 0)
    goto pii_error3;
  i = imp_func_tab_init (&((*pspm)->func_unknowns));
  if (i != 0)
    goto pii_error3;

  return 0;

pii_error3:
  module_release ((*pspm)->module);
  pspm_imp_free (*pspm);
  sfree (*pspm);
  return -1;
pii_error2:
  module_release ((*pspm)->module);
  module_free ((*pspm)->module);
  sfree ((*pspm)->module);
pii_error1:
  sfree (*pspm);
  return -1;
}

void
pspm_imp_free (pspm_imp_t * pspm)
{
  imp_plugin_t *p;
  sipevent_t *evt;

  if (pspm == NULL)
    return;

  imp_func_tab_free (pspm->func_invites);
  imp_func_tab_free (pspm->func_acks);
  imp_func_tab_free (pspm->func_registers);
  imp_func_tab_free (pspm->func_byes);
  imp_func_tab_free (pspm->func_optionss);
  imp_func_tab_free (pspm->func_infos);
  imp_func_tab_free (pspm->func_cancels);
  imp_func_tab_free (pspm->func_notifys);
  imp_func_tab_free (pspm->func_subscribes);
  imp_func_tab_free (pspm->func_unknowns);

  sfree (pspm->func_invites);
  sfree (pspm->func_acks);
  sfree (pspm->func_registers);
  sfree (pspm->func_byes);
  sfree (pspm->func_optionss);
  sfree (pspm->func_infos);
  sfree (pspm->func_cancels);
  sfree (pspm->func_notifys);
  sfree (pspm->func_subscribes);
  sfree (pspm->func_unknowns);

  module_free (pspm->module);
  sfree (pspm->module);

  evt = fifo_tryget (pspm->sip_traffic);
  while (evt != NULL)
    {
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      evt = fifo_tryget (pspm->sip_traffic);
    }

  fifo_free (pspm->sip_traffic);
  sfree (pspm->sip_traffic);

  for (p = pspm->imp_plugins; p != NULL; p = pspm->imp_plugins)
    {
      REMOVE_ELEMENT (pspm->imp_plugins, p);
      imp_plugin_free (p);
      sfree (p);
    }

}

int
pspm_imp_release (pspm_imp_t * pspm)
{
  int i;

  i = module_release (pspm->module);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_imp_start (pspm_imp_t * pspm, void *(*func_start) (void *), void *arg)
{
  int i;

  i = module_start (pspm->module, func_start, arg);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_imp_execute (pspm_imp_t * pspm, int sec_max, int usec_max, int max)
{
  fd_set imp_fdset;
  int max_fd;
  struct timeval tv;
  sipevent_t *evt;

  while (1)
    {
      int s;
      int i;

      tv.tv_sec = sec_max;
      tv.tv_usec = usec_max;

      max--;
      FD_ZERO (&imp_fdset);
      s = ppl_pipe_get_read_descr (pspm->module->wakeup);
      if (s <= 1)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "No wakeup value in module imp!\n"));
          return -1;
        }
      FD_SET (s, &imp_fdset);

      max_fd = s;
      /* It could be interesting to use sockets to wake up plugins?
         This can be done like in the tlp module. */
      /*
         for (plug=pspm->plugins; plug!=NULL; plug=plug->next)
         {
         if (plug->socket>0)
         {
         if (max_fd<plug->socket) max_fd = plug->socket;
         FD_SET(plug->socket, &imp_fdset);
         }
         }
       */
      if ((sec_max == -1) || (usec_max == -1))
        i = select (max_fd + 1, &imp_fdset, NULL, NULL, NULL);
      else
        i = select (max_fd + 1, &imp_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &imp_fdset))
        {                       /* time to wake up: we receive "q" to quit and any other char to read all socket */
          char tmp[2];

          i = ppl_pipe_read (pspm->module->wakeup, tmp, 1);
          if (i == 1 && tmp[0] == 'q')
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_INFO2, NULL,
                           "Exiting tlp module!\n"));
              return 0;
            }
        }

      evt = fifo_tryget (pspm->sip_traffic);
      while (evt != NULL)
        {
          pspm_imp_dispatch_traffic (pspm, evt);
          evt = fifo_tryget (pspm->sip_traffic);
        }

      if (max == 0)
        return 0;
    }
}

int
pspm_imp_dispatch_traffic (pspm_imp_t * pspm, sipevent_t * evt)
{
  psp_req_t *psp_req;
  int i;

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "imp module: dispatching request!\n"));


  /* search for an existing transaction (we search one before, but
     it could be usefull in case the system is overloaded. */
  i = psp_core_find_transaction_and_add_event (evt);
  if (i == 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Find transaction for event. (Anyway, this should not happen here)!\n"));
      return 0;
    }

  /* so it's a new request. (it must be?) */
  /* it's time to create a psp_req_t element and call plugins */

  i = psp_req_init (&psp_req, evt->sip);
  if (i != 0)
    {
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Could not create psp_req!\n"));
      return -1;
    }
  sfree (evt);
  i = pspm_imp_call_plugins (pspm, psp_req);
  if (i == -2)
    {
      psp_req_free (psp_req);
      sfree (psp_req);
      return -1;
    }
  i = pspm_imp_dispatch_psp_request (pspm, psp_req);
  if (i != 0)
    return -1;                  /* psp_req is release inside pspm_imp_dispatch_psp_request()
                                   if needed */
  return 0;
}

int
pspm_imp_dispatch_psp_request (pspm_imp_t * pspm, psp_req_t * req)
{
  /* Plugins have been called and the request should not be
     sent to a module: UAP, SLP or SFP */
  /* case1: flag in req->flag contains the bit for UAS_MODE */
  if (__IS_PSP_UAS_MODE (req->flag))
    {
      int i;
      transaction_t *transaction;
      osip_ctx_t *osip_ctx;
      sipevent_t *evt;
      sip_t *request;

      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Swithing to uas mode!\n"));
      /* special ACK case */
      /* if no transaction has been found, and if the ACK is coming here,
         this means it's a late ACK. This is not an issue
         as in this case, the ACK is probably for a transaction that
         previously failed. (else the request sould contain enough
         informations to be forwarded and should be targetted to
         the uap module!) */
      if (MSG_IS_ACK (req->request))
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_WARNING, NULL,
                       "late ACK (probably for a bad transaction) is discarded!\n"));
          psp_req_free (req);
          sfree (req);
          return 0;
        }

      /* I PREFER TO CLONE THE REQUEST, but this may be avoided if
         we set psp_req->request = NULL; before releasing psp_req
         in module uap... I prefer to leave it as a possible improvement
         when the proxy will become stable... */
      i = msg_clone (req->request, &request);
      if (i != 0)
        {
          psp_req_free (req);
          sfree (req);
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "Could not clone request!\n"));
          return -1;
        }

      evt = osip_new_incoming_sipmessage (request);
      transaction = psp_core_uap_create_transaction (evt);

      if (transaction == NULL)
        {                       /* ?? plugin should be informed that this context failed */
          msg_free (evt->sip);
          sfree (evt->sip);
          sfree (evt);
          psp_req_free (req);
          sfree (req);
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "Could not create a transaction!\n"));
          return -1;
        }
      fifo_add (transaction->transactionff, evt);

      psp_req_set_osip_id (req, transaction->transactionid);

      osip_ctx_init (&osip_ctx, transaction);
      osip_ctx_start (osip_ctx);

      psp_core_event_add_uas_request (req);
      return 0;
    }
  if (__IS_PSP_SLESS_MODE (req->flag))
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Swithing to sless mode!\n"));
      psp_core_event_add_sless_request (req);
      return 0;
    }
  if (__IS_PSP_SFULL_MODE (req->flag))
    {
      /* 2: forward the psp_req element to sfp */
      psp_core_event_add_sfull_request (req);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Swithing to sfull mode!\n"));
      return 0;
    }
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_BUG, NULL,
               "BUG: No mode asked?? deleting request!\n"));
  psp_req_free (req);
  sfree (req);
  return -1;
}

int
pspm_imp_call_plugins (pspm_imp_t * pspm, psp_req_t * req)
{
  imp_func_tab_t *t;
  imp_func_t *f;

  if (MSG_IS_INVITE (req->request))
    t = pspm->func_invites;
  else if (MSG_IS_ACK (req->request))
    t = pspm->func_acks;
  else if (MSG_IS_REGISTER (req->request))
    t = pspm->func_registers;
  else if (MSG_IS_BYE (req->request))
    t = pspm->func_byes;
  else if (MSG_IS_OPTIONS (req->request))
    t = pspm->func_optionss;
  else if (MSG_IS_INFO (req->request))
    t = pspm->func_infos;
  else if (MSG_IS_CANCEL (req->request))
    t = pspm->func_cancels;
  else if (MSG_IS_NOTIFY (req->request))
    t = pspm->func_notifys;
  else if (MSG_IS_SUBSCRIBE (req->request))
    t = pspm->func_subscribes;
  else
    t = pspm->func_unknowns;

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_rcv_func (req);
      if (__IS_PSP_MANDATE (req->flag))
        return 0;
      if (__IS_PSP_STOP (req->flag))
        return -2;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_rcv_func (req);
      if (__IS_PSP_MANDATE (req->flag))
        return 0;
      if (__IS_PSP_STOP (req->flag))
        return -2;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_rcv_func (req);
      if (__IS_PSP_MANDATE (req->flag))
        return 0;
      if (__IS_PSP_STOP (req->flag))
        return -2;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_rcv_func (req);
      if (__IS_PSP_MANDATE (req->flag))
        return 0;
      if (__IS_PSP_STOP (req->flag))
        return -2;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_rcv_func (req);
      if (__IS_PSP_MANDATE (req->flag))
        return 0;
      if (__IS_PSP_STOP (req->flag))
        return -2;
    }
  return 0;
}
