/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

#include <osip_ctx.h>

#ifdef OSIP_MT

void *
transaction_thread (transaction_t * transaction)
{
  sipevent_t *se;

#ifdef __VXWORKS_OS__
  taskSafe ();                  /* prevent from deletion */
#endif
  while (1)
    {
      se = (sipevent_t *) fifo_get (transaction->transactionff);
      if (se == NULL)
        {
          return NULL;
        }
      if (transaction_execute (transaction, se) < 1)    /* deletion asked */
        {
#ifdef __VXWORKS_OS__
          taskUnsafe ();        /* thread is safe for deletion */
#endif
          sthread_exit ();
#ifdef WIN32
          {
            osip_ctx_t *ctx;

            ctx = (osip_ctx_t *) (transaction->your_instance);
            CloseHandle (ctx->thread->h);
          }
#endif
          return NULL;
        }
    }
}


int
osip_ctx_init (osip_ctx_t ** osip_ctx, transaction_t * transaction)
{
  *osip_ctx = (osip_ctx_t *) smalloc (sizeof (osip_ctx_t));
  if (*osip_ctx == NULL)
    return -1;
  (*osip_ctx)->transaction = transaction;
  (*osip_ctx)->thread = NULL;
  return 0;
}

int
osip_ctx_start (osip_ctx_t * osip_ctx)
{
  osip_ctx->transaction->your_instance = osip_ctx;

  osip_ctx->thread =
    sthread_create (20000, NULL, (void *(*)(void *)) transaction_thread,
                    (void *) osip_ctx->transaction);
  /* try to give a highest priority to this task     */
  /* this is only supported by some OS and currently */
  /* only used on VxWorks. */
  if (osip_ctx->thread == NULL)
    return -1;
  sthread_setpriority (osip_ctx->thread, 1);
  return 0;
}

void
osip_ctx_stop (osip_ctx_t * osip_ctx)
{
  if (osip_ctx->thread != NULL)
    {
      psp_core_t *psp_core;
      sipevent_t *sipevent;

      sipevent =
        osip_new_event (KILL_TRANSACTION, osip_ctx->transaction->transactionid);
      fifo_add (osip_ctx->transaction->transactionff, sipevent);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "Add a KILL_TRANSACTION event!\n"));

      /* Another thread must call pthread_join() */
      psp_core = psp_core_get ();
      fifo_add (psp_core->psp_osip->osip_tr_terminated, osip_ctx);
  } else
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_BUG, NULL,
                   "How can I be there??? thread is broken!\n"));
    }
}

#endif /* OSIP_MT */
