/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>
#include <ppl/ppl_dns.h>

extern psp_core_t *core;

PPL_DECLARE (int)
psp_core_dns_add_domain_error (char *address)
{
  ppl_dns_error_add (address);
  module_wakeup (core->sfp->module);    /* this module needs to be waken up */
  return 0;
}

PPL_DECLARE (int)
psp_core_dns_add_domain_result (ppl_dns_entry_t * dns)
{
  ppl_dns_add_domain_result (dns);
  module_wakeup (core->sfp->module);    /* this module needs to be waken up */
  return 0;
}


int
psp_core_event_is_for_stateless (sipevent_t * evt)
{
  via_t *via;
  generic_param_t *b;

  /* partysip generate 2 kinds of branches:
     The 8th char "f" for statefull answers
     and "l" for stateless response.  :-/ */
  via = list_get (evt->sip->vias, 0);
  via_param_getbyname (via, "branch", &b);
  if (b == NULL)                /* impossible... response is discarded... */
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "discard response with Not acceptable Via!\n"));
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "This response may be for another proxy?!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      return -2;                /* STOP PROCESSING */
    }
  if (strlen (b->gvalue) < 8)   /* impossible... response is discarded... */
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "discard response with Not acceptable Via!\n"));
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "This response may be for another proxy?!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      return -2;
    }

  /* STATELESS */
  if (0 == strncmp (b->gvalue + 7, "l", 1))
    {                           /* this response is for a stateless context! */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "imp module: This response is for the stateless module!\n"));
      psp_core_event_add_sless_response (evt->sip);
      sfree (evt);
      return 0;
    }
  return -1;
}


PPL_DECLARE (int)
psp_core_event_resolv_add_query (char *address)
{
  ppl_dns_add_domain_query (address);
  module_wakeup (core->resolv->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_imp_traffic (sipevent_t * req)
{
  fifo_add (core->imp->sip_traffic, req);
  module_wakeup (core->imp->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_sip_message (sipevent_t * evt)
{
  int i;

  /* be sure, we won't crash in strncmp */
  if (MSG_IS_RESPONSE (evt->sip))
    {
      i = psp_core_event_is_for_stateless (evt);
      if (i == -2)              /* error */
        return -1;
      else if (i == 0)
        return 0;
      /* i==-1, we have to continue searching for a responsible module. */

    }
  i = psp_core_find_transaction_and_add_event (evt);
  if (i == 0)
    return 0;

  if (MSG_IS_REQUEST (evt->sip))
    {
      /* the slp module will check himself if this request is for him */
      if (MSG_IS_ACK (evt->sip))
        {                       /* continue as it is a new transaction... */
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_INFO3, NULL,
                       "This is a ACK for INVITE!\n"));
          /* imp is responsible of that one */
          psp_core_event_add_imp_traffic (evt);
          return 0;
      } else if (MSG_IS_CANCEL (evt->sip))
        {                       /* continue as it is a new transaction... */
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_INFO3, NULL,
                       "This is a CANCEL for INVITE (propably for an INVITE.)!\n"));
          psp_core_event_add_imp_traffic (evt);
          return 0;
        }
      psp_core_event_add_slp_traffic (evt);
      /* slp module will decide if he owns the request or not. */
      return 0;
  } else
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "imp module: No module seems to be able to forward this response!\n"));
      /* this is probably a late answer? */
      /* let's forward it! */
      i = psp_core_handle_late_answer (evt->sip);
      if (i != 0)
        {
          msg_free (evt->sip);
          sfree (evt->sip);
          sfree (evt);
          return -1;
        }
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
    }
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_uas_request (psp_req_t * req)
{
  fifo_add (core->uap->uas_request, req);
  module_wakeup (core->uap->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_slp_traffic (sipevent_t * evt)
{
  fifo_add (core->slp->sless_traffic, evt);
  module_wakeup (core->slp->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_sless_request (psp_req_t * req)
{
  fifo_add (core->slp->sless_request, req);
  module_wakeup (core->slp->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_sless_response (sip_t * resp)
{
  fifo_add (core->slp->sless_response, resp);
  module_wakeup (core->slp->module);
  return 0;
}

PPL_DECLARE (int)
psp_core_event_add_sfull_request (psp_req_t * req)
{
  fifo_add (core->sfp->sfull_request, req);
  module_wakeup (core->sfp->module);
  return 0;
}

/*
int  psp_core_event_ask_sfp(psp_req_t *req)
{
  fifo_add(core->sfp->sfull_traffic, req);
  module_wakeup(core->sfp->module);
  return 0;
}
*/
PPL_DECLARE (int)
psp_core_event_add_sfull_response (psp_req_t * req)
{
  fifo_add (core->sfp->sfull_response, req);
  module_wakeup (core->sfp->module);
  return 0;
}
