/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

#include <osip_ctx.h>

void
psp_core_cb_endoftransaction (transaction_t * tr)
{
#ifdef OSIP_MT
  osip_ctx_stop ((osip_ctx_t *) tr->your_instance);
#endif
}

void
psp_core_cb_ist_kill_transaction (transaction_t * tr)
{
  static int ist_killed = 0;

  ist_killed++;
  if (ist_killed % 100 == 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Number of killed IST transaction = %i\n", ist_killed));
    }
#ifdef OSIP_MT
  osip_ctx_stop ((osip_ctx_t *) tr->your_instance);
#endif
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_ist_kill_transaction!\n"));
}

void
psp_core_cb_nist_kill_transaction (transaction_t * tr)
{
  static int nist_killed = 0;

  nist_killed++;
  if (nist_killed % 100 == 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Number of killed NIST transaction = %i\n", nist_killed));
    }
#ifdef OSIP_MT
  osip_ctx_stop ((osip_ctx_t *) tr->your_instance);
#endif
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_nist_kill_transaction!\n"));
}

void
psp_core_cb_sndresp_retransmission (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_sndresp_retransmission!\n"));
}

void
psp_core_cb_rcvreq_retransmission (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_rcvreq_retransmission!\n"));
}

void
psp_core_cb_rcvinvite (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvinvite!\n"));
}

void
psp_core_cb_rcvack (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvack!\n"));
}

void
psp_core_cb_rcvack2 (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvack2!\n"));
}

void
psp_core_cb_rcvregister (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_rcvregister!\n"));
}

void
psp_core_cb_rcvbye (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvbye!\n"));
}

void
psp_core_cb_rcvcancel (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvcancel!\n"));
}

void
psp_core_cb_rcvinfo (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvinfo!\n"));
}

void
psp_core_cb_rcvoptions (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvoptions!\n"));
}

void
psp_core_cb_rcvprack (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcvprack!\n"));
}

void
psp_core_cb_rcvunkrequest (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_rcvunkrequest!\n"));
}

void
psp_core_cb_snd1xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd1xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {

  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }
}

void
psp_core_cb_snd2xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd2xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }

}

void
psp_core_cb_snd3xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd3xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }
}

void
psp_core_cb_snd4xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd4xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }
}

void
psp_core_cb_snd5xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd5xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }

}

void
psp_core_cb_snd6xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_snd6xx!\n"));
  if (MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "REGISTER"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "BYE"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "OPTIONS"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "INFO"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "CANCEL"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "NOTIFY"))
    {
  } else if (MSG_IS_RESPONSEFOR (sip, "SUBSCRIBE"))
    {
  } else
    {
    }
}
