/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <partysip.h>

#include <osip_ctx.h>

void
psp_core_cb_ict_kill_transaction (transaction_t * tr)
{
  static int ict_killed = 0;

  ict_killed++;
  if (ict_killed % 100 == 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Number of killed ICT transaction = %i\n", ict_killed));
    }
#ifdef OSIP_MT
  osip_ctx_stop ((osip_ctx_t *) tr->your_instance);
#endif
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_ict_kill_transaction!\n"));
}

void
psp_core_cb_nict_kill_transaction (transaction_t * tr)
{
  static int nict_killed = 0;

  nict_killed++;
  if (nict_killed % 100 == 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "Number of killed NICT transaction = %i\n", nict_killed));
    }
#ifdef OSIP_MT
  osip_ctx_stop ((osip_ctx_t *) tr->your_instance);
#endif
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_nict_kill_transaction!\n"));
}

void
psp_core_cb_transport_error (transaction_t * tr, int error)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_ERROR, NULL,
               "psp_core_cb_network_error! \n"));
}

void
psp_core_cb_sndreq_retransmission (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_sndreq_retransmission! \n"));
}

void
psp_core_cb_rcvresp_retransmission (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_rcvresp_retransmission! \n"));
}

void
psp_core_cb_sndinvite (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndinvite!\n"));
}

void
psp_core_cb_sndack (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndack!\n"));
}

void
psp_core_cb_sndbye (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndbye!\n"));
}

void
psp_core_cb_sndcancel (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndcancel!\n"));
}

void
psp_core_cb_sndinfo (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndinfo!\n"));
}

void
psp_core_cb_sndoptions (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndoptions!\n"));
}

void
psp_core_cb_sndregister (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_sndregister!\n"));
}

void
psp_core_cb_sndprack (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_sndprack!\n"));
}

void
psp_core_cb_sndunkrequest (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "psp_core_cb_sndunkrequest!\n"));
}

void
psp_core_cb_rcv1xx (transaction_t * tr, sip_t * sip)
{
  psp_req_t *req;
  sfp_branch_t *sfp;
  sip_t *response;
  int i;
  psp_core_t *core = psp_core_get ();

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv1xx!\n"));

  pspm_sfp_lock_sfp_pending_access (core->sfp);
  req = psp_core_sfp_find_req (tr->transactionid);
  if (req == NULL)
    {
      pspm_sfp_unlock_sfp_pending_access (core->sfp);
      return;
    }
  sfp = psp_core_sfp_find_branch (req, tr->transactionid);
  if (sfp == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "The answer is too late..!\n"));
      pspm_sfp_unlock_sfp_pending_access (core->sfp);
      return;
    }
  if (sfp->response == NULL)
    {
      i = msg_clone (sip, &response);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "Could not clone response1xx!\n"));
          pspm_sfp_unlock_sfp_pending_access (core->sfp);
          return;
        }
      if (!MSG_IS_RESPONSEFOR (response, "INVITE"))
        {
          psp_req_set_timeout (req, 32);
          sfp_branch_set_timeout (sfp, 32);
      } else
        {
          psp_req_set_timeout (req, 160);
          sfp_branch_set_timeout (sfp, 160);
        }
      sfp_branch_set_response (sfp, response);
      psp_core_event_add_sfull_response (req);
  } else if (0 ==
             strcmp (sip->strtline->statuscode,
                     sfp->response->strtline->statuscode))
    {                           /* discard useless event! */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "Discarding useless event for sfp!\n"));
  } else if (0 >
             strcmp (sfp->response->strtline->statuscode,
                     sip->strtline->statuscode))
    {                           /* discard useless event! */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "Discarding useless event for sfp!\n"));
  } else                        /* updated answer code! */
    {
      /* this MUST be a response for the statefull module */
      i = msg_clone (sip, &response);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "Could not clone response1xx!\n"));
          pspm_sfp_unlock_sfp_pending_access (core->sfp);
          return;
        }
      msg_free (sfp->response);
      sfree (sfp->response);

      sfp_branch_set_response (sfp, response);
      psp_core_event_add_sfull_response (req);
    }

  pspm_sfp_unlock_sfp_pending_access (core->sfp);
}

void
psp_core_distribute_to_sfp_module (transaction_t * tr, sip_t * sip)
{
  psp_req_t *req;
  sfp_branch_t *sfp;
  sip_t *response;
  int i;
  psp_core_t *core = psp_core_get ();

  pspm_sfp_lock_sfp_pending_access (core->sfp);
  req = psp_core_sfp_find_req (tr->transactionid);
  if (req == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "psp_core_cb_rcv23456xx No sfp transaction!\n"));
      pspm_sfp_unlock_sfp_pending_access (core->sfp);
      return;
    }

  sfp = psp_core_sfp_find_branch (req, tr->transactionid);

  if (sfp == NULL)
    {
      pspm_sfp_unlock_sfp_pending_access (core->sfp);
      return;
    }
  /*  if (sfp->response==NULL) { */

  /* even if we can't go further, this transaction and request will timetout
     soon */
  if (!MSG_IS_RESPONSEFOR (sip, "INVITE"))
    {
      psp_req_set_timeout (req, 5);
      sfp_branch_set_timeout (sfp, 5);
  } else
    {
      psp_req_set_timeout (req, 5);
      sfp_branch_set_timeout (sfp, 5);
    }

  i = msg_clone (sip, &response);
  if (i != 0)
    {
      pspm_sfp_unlock_sfp_pending_access (core->sfp);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Could not clone response1xx!\n"));
      return;
    }
  /* this method is only called once, so we are sure that
     sfp->response is an informational answer */
  if (sfp->response != NULL)
    {
      msg_free (sfp->response);
      sfree (sfp->response);
    }
  sfp_branch_set_response (sfp, response);
  pspm_sfp_unlock_sfp_pending_access (core->sfp);
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "psp_core_distribute_to_sfp_module before adding statefull req!\n"));
  psp_core_event_add_sfull_response (req);
}

void
psp_core_cb_rcv2xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv2xx!\n"));
  /* this MUST be a response for the statefull module */
  psp_core_distribute_to_sfp_module (tr, sip);
}

void
psp_core_cb_rcv3xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv3xx!\n"));
  /* this MUST be a response for the statefull module */
  psp_core_distribute_to_sfp_module (tr, sip);
}

void
psp_core_cb_rcv4xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv4xx!\n"));
  psp_core_distribute_to_sfp_module (tr, sip);
}

void
psp_core_cb_rcv5xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv5xx!\n"));
  psp_core_distribute_to_sfp_module (tr, sip);
}

void
psp_core_cb_rcv6xx (transaction_t * tr, sip_t * sip)
{
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL, "psp_core_cb_rcv6xx!\n"));
  psp_core_distribute_to_sfp_module (tr, sip);
}
