/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>

#include <osip_ctx.h>


void *
osip_timers_thread (psp_osip_t * psp_osip)
{
  osip_ctx_t *osip_ctx;
  int i;

  while (1)
    {
      susleep (psp_osip->delay * 1000);
      osip_timers_ict_execute (psp_osip->osip);
      osip_timers_ist_execute (psp_osip->osip);
      osip_timers_nict_execute (psp_osip->osip);
      osip_timers_nist_execute (psp_osip->osip);

      osip_ctx = fifo_tryget (psp_osip->osip_tr_terminated);
      while (osip_ctx != NULL)
        {
          i = sthread_join (osip_ctx->thread);
          if (i != 0)
            {
              fprintf (stderr, "Cannot join thread!! error code: %i\n", i);
          } else
            {
              transaction_free (osip_ctx->transaction);
              sfree (osip_ctx->transaction);
              sfree (osip_ctx->thread);
              sfree (osip_ctx);
            }
          osip_ctx = fifo_tryget (psp_osip->osip_tr_terminated);
        }
      if (psp_osip->exit_flag == 1)
        return NULL;
    }
  return 0;
}

int
psp_osip_init (psp_osip_t ** psp_osip, int delay)
{
  osip_t *osip;
  int i;

  /* initialize osip */
  if (-1 == osip_global_init ())
    return -1;                  /* mutex is not initialised properly */

  i = osip_init (&osip);
  if (i != 0)
    return -1;

  osip_setcb_send_message (osip, &psp_core_cb_snd_message);

  osip_setcb_ict_kill_transaction (osip, &psp_core_cb_ict_kill_transaction);
  osip_setcb_ist_kill_transaction (osip, &psp_core_cb_ist_kill_transaction);
  osip_setcb_nict_kill_transaction (osip, &psp_core_cb_nict_kill_transaction);
  osip_setcb_nist_kill_transaction (osip, &psp_core_cb_nist_kill_transaction);

  osip_setcb_ict_2xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_ict_3456xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_ict_invite_sent2 (osip, &psp_core_cb_sndreq_retransmission);
  osip_setcb_ist_2xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_ist_3456xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_ist_invite_received2 (osip, &psp_core_cb_rcvreq_retransmission);
  osip_setcb_nict_2xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_nict_3456xx_received2 (osip, &psp_core_cb_rcvresp_retransmission);
  osip_setcb_nict_request_sent2 (osip, &psp_core_cb_sndreq_retransmission);
  osip_setcb_nist_2xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_nist_3456xx_sent2 (osip, &psp_core_cb_sndresp_retransmission);
  osip_setcb_nist_request_received2 (osip, &psp_core_cb_rcvreq_retransmission);

  /*  osip_setcb_killtransaction(osip,&psp_core_cb_killtransaction);
     osip_setcb_endoftransaction(osip,&psp_core_cb_endoftransaction); */

  osip_setcb_ict_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_ist_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_nict_transport_error (osip, &psp_core_cb_transport_error);
  osip_setcb_nist_transport_error (osip, &psp_core_cb_transport_error);

  osip_setcb_ict_invite_sent (osip, &psp_core_cb_sndinvite);
  osip_setcb_ict_ack_sent (osip, &psp_core_cb_sndack);
  osip_setcb_nict_register_sent (osip, &psp_core_cb_sndregister);
  osip_setcb_nict_bye_sent (osip, &psp_core_cb_sndbye);
  osip_setcb_nict_cancel_sent (osip, &psp_core_cb_sndcancel);
  osip_setcb_nict_info_sent (osip, &psp_core_cb_sndinfo);
  osip_setcb_nict_options_sent (osip, &psp_core_cb_sndoptions);
  osip_setcb_nict_subscribe_sent (osip, &psp_core_cb_sndoptions);
  osip_setcb_nict_notify_sent (osip, &psp_core_cb_sndoptions);
  /*  osip_setcb_nict_sndprack   (osip,&psp_core_cb_sndprack); */
  osip_setcb_nict_unknown_sent (osip, &psp_core_cb_sndunkrequest);

  osip_setcb_ict_1xx_received (osip, &psp_core_cb_rcv1xx);
  osip_setcb_ict_2xx_received (osip, &psp_core_cb_rcv2xx);
  osip_setcb_ict_3xx_received (osip, &psp_core_cb_rcv3xx);
  osip_setcb_ict_4xx_received (osip, &psp_core_cb_rcv4xx);
  osip_setcb_ict_5xx_received (osip, &psp_core_cb_rcv5xx);
  osip_setcb_ict_6xx_received (osip, &psp_core_cb_rcv6xx);

  osip_setcb_ist_1xx_sent (osip, &psp_core_cb_snd1xx);
  osip_setcb_ist_2xx_sent (osip, &psp_core_cb_snd2xx);
  osip_setcb_ist_3xx_sent (osip, &psp_core_cb_snd3xx);
  osip_setcb_ist_4xx_sent (osip, &psp_core_cb_snd4xx);
  osip_setcb_ist_5xx_sent (osip, &psp_core_cb_snd5xx);
  osip_setcb_ist_6xx_sent (osip, &psp_core_cb_snd6xx);

  osip_setcb_nict_1xx_received (osip, &psp_core_cb_rcv1xx);
  osip_setcb_nict_2xx_received (osip, &psp_core_cb_rcv2xx);
  osip_setcb_nict_3xx_received (osip, &psp_core_cb_rcv3xx);
  osip_setcb_nict_4xx_received (osip, &psp_core_cb_rcv4xx);
  osip_setcb_nict_5xx_received (osip, &psp_core_cb_rcv5xx);
  osip_setcb_nict_6xx_received (osip, &psp_core_cb_rcv6xx);

  osip_setcb_nist_1xx_sent (osip, &psp_core_cb_snd1xx);
  osip_setcb_nist_2xx_sent (osip, &psp_core_cb_snd2xx);
  osip_setcb_nist_3xx_sent (osip, &psp_core_cb_snd3xx);
  osip_setcb_nist_4xx_sent (osip, &psp_core_cb_snd4xx);
  osip_setcb_nist_5xx_sent (osip, &psp_core_cb_snd5xx);
  osip_setcb_nist_6xx_sent (osip, &psp_core_cb_snd6xx);

  osip_setcb_ist_invite_received (osip, &psp_core_cb_rcvinvite);
  osip_setcb_ist_ack_received (osip, &psp_core_cb_rcvack);
  osip_setcb_ist_ack_received2 (osip, &psp_core_cb_rcvack2);
  osip_setcb_nist_register_received (osip, &psp_core_cb_rcvregister);
  osip_setcb_nist_bye_received (osip, &psp_core_cb_rcvbye);
  osip_setcb_nist_cancel_received (osip, &psp_core_cb_rcvcancel);
  osip_setcb_nist_info_received (osip, &psp_core_cb_rcvinfo);
  osip_setcb_nist_options_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_subscribe_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_notify_received (osip, &psp_core_cb_rcvoptions);
  osip_setcb_nist_unknown_received (osip, &psp_core_cb_rcvunkrequest);

  (*psp_osip) = (psp_osip_t *) smalloc (sizeof (psp_osip_t));
  if ((*psp_osip) == NULL)
    goto poi_error0;

  (*psp_osip)->osip = osip;

  if (100 < delay < 5000)
    (*psp_osip)->delay = delay;
  else
    delay = 500;

  (*psp_osip)->exit_flag = 0;   /* 0 not set, 1 set */
  (*psp_osip)->timers = NULL;

  (*psp_osip)->osip_tr_terminated = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*psp_osip)->osip_tr_terminated == NULL)
    goto poi_error1;
  fifo_init ((*psp_osip)->osip_tr_terminated);

  (*psp_osip)->timers =
    sthread_create (20000, NULL, (void *(*)(void *)) osip_timers_thread,
                    (void *) (*psp_osip));

  if ((*psp_osip)->timers == NULL)
    goto poi_error2;

  return 0;

poi_error2:
  fifo_free ((*psp_osip)->osip_tr_terminated);
  sfree ((*psp_osip)->osip_tr_terminated);
poi_error1:
  (*psp_osip)->exit_flag = 1;
  psp_osip_free (*psp_osip);
  sfree (*psp_osip);
  return -1;
poi_error0:
  osip_free (osip);
  sfree (osip);
  return -1;
}

void
_osip_kill_transaction (list_t * transactions)
{
  transaction_t *transaction;
  osip_ctx_t *osip_ctx;
  int i;
  int pos = 0;

  while (!list_eol (transactions, pos))
    {
      transaction = list_get (transactions, pos);
      osip_ctx = transaction->your_instance;

      if (osip_ctx->thread != NULL)
        {
          sipevent_t *sipevent;

          sipevent = osip_new_event (KILL_TRANSACTION, transaction->transactionid);
          fifo_add (transaction->transactionff, sipevent);

          i = sthread_join (osip_ctx->thread);
          if (i != 0)
            printf ("shutting down thread (XXXX) : ERROR\n");
          else
            {
#ifdef __sun__
              printf ("shutting down thread (%ul) : done\n", *osip_ctx->thread);
#elif defined(__OpenBSD__)
              printf ("shutting down threadx : done\n");
#else
              printf ("shutting down thread (%li) : done\n", *osip_ctx->thread);
#endif
              sfree (osip_ctx->thread);
              sfree (osip_ctx);
              transaction_free (transaction);
              sfree (transaction);
            }
      } else
        {
          pos++;
          printf ("I can't do that? (unknown error)\n");
        }
    }
}

void
psp_osip_free (psp_osip_t * psp_osip)
{
  osip_ctx_t *osip_ctx;
  int i;

  if (psp_osip == NULL)
    return;


  psp_osip->exit_flag = 1;      /* this value is checked by the timer thread */
  if (psp_osip->timers != NULL)
    {
      i = sthread_join (psp_osip->timers);
      if (i != 0)
        printf ("ERROR: can't stop timer thread\n");
      sfree (psp_osip->timers);
      psp_osip->timers = NULL;
    }

  /* now it's time to free all transactions... */
  _osip_kill_transaction (psp_osip->osip->ict_transactions);
  _osip_kill_transaction (psp_osip->osip->nict_transactions);
  _osip_kill_transaction (psp_osip->osip->ist_transactions);
  _osip_kill_transaction (psp_osip->osip->nist_transactions);

  /* Some elements (!!already released!!) might still be there. */
  osip_ctx = fifo_tryget (psp_osip->osip_tr_terminated);
  while (osip_ctx != NULL)
    {
      osip_ctx = fifo_tryget (psp_osip->osip_tr_terminated);
    }
  fifo_free (psp_osip->osip_tr_terminated);
  sfree (psp_osip->osip_tr_terminated);

  osip_free (psp_osip->osip);
  sfree (psp_osip->osip);
  osip_global_free ();
}


int
psp_osip_release (psp_osip_t * psp_osip)
{
  int i;

  if (psp_osip == NULL)
    return -1;
  psp_osip->exit_flag = 1;
  if (psp_osip->timers != NULL)
    {
      i = sthread_join (psp_osip->timers);
      if (i != 0)
        {
          psp_osip->timers = NULL;
          printf ("ERROR: can't stop timer thread\n");
          return -1;
      } else
        sfree (psp_osip->timers);
    }
  psp_osip->timers = NULL;
  return 0;
}

int
psp_osip_set_delay (psp_osip_t * psp_osip, int delay)
{
  if (psp_osip == NULL)
    return -1;
  psp_osip->delay = delay;
  return 0;
}
