/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip/partysip.h>

#include <partysip/osip_ctx.h>
#include <partysip/osip_msg.h>

#include <ppl/ppl_dns.h>

int
pspm_sfp_init (pspm_sfp_t ** pspm)
{
  int i;

  (*pspm) = (pspm_sfp_t *) smalloc (sizeof (pspm_sfp_t));
  if (*pspm == NULL)
    return -1;

  i = module_init (&((*pspm)->module), MOD_THREAD, "sfp");
  if (i != 0)
    goto psi_error1;
  (*pspm)->sfp_plugins = NULL;


  (*pspm)->mut_sfp_pending = smutex_init ();
  if ((*pspm)->mut_sfp_pending == NULL)
    goto psi_error2;
  (*pspm)->sfp_pendings = NULL;

  (*pspm)->sfull_request = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sfull_request == NULL)
    goto psi_error3;
  fifo_init ((*pspm)->sfull_request);

  (*pspm)->sfull_response = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sfull_response == NULL)
    goto psi_error4;
  fifo_init ((*pspm)->sfull_response);

  (*pspm)->req_terminated = NULL;

  (*pspm)->fwd_invites = NULL;
  (*pspm)->fwd_acks = NULL;
  (*pspm)->fwd_registers = NULL;
  (*pspm)->fwd_byes = NULL;
  (*pspm)->fwd_optionss = NULL;
  (*pspm)->fwd_infos = NULL;
  (*pspm)->fwd_cancels = NULL;
  (*pspm)->fwd_notifys = NULL;
  (*pspm)->fwd_subscribes = NULL;
  (*pspm)->fwd_unknowns = NULL;

  (*pspm)->rcv_1xxs = NULL;
  (*pspm)->rcv_2xxs = NULL;
  (*pspm)->rcv_3xxs = NULL;
  (*pspm)->rcv_4xxs = NULL;
  (*pspm)->rcv_5xxs = NULL;
  (*pspm)->rcv_6xxs = NULL;

  (*pspm)->snd_1xxs = NULL;
  (*pspm)->snd_2xxs = NULL;
  (*pspm)->snd_3xxs = NULL;
  (*pspm)->snd_4xxs = NULL;
  (*pspm)->snd_5xxs = NULL;
  (*pspm)->snd_6xxs = NULL;

  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_invites));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_acks));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_registers));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_byes));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_optionss));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_infos));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_cancels));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_notifys));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_subscribes));
  if (i != 0)
    goto psi_error5;
  i = sfp_fwd_func_tab_init (&((*pspm)->fwd_unknowns));
  if (i != 0)
    goto psi_error5;

  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_1xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_2xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_3xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_4xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_5xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_rcv_func_tab_init (&((*pspm)->rcv_6xxs));
  if (i != 0)
    goto psi_error5;

  i = sfp_snd_func_tab_init (&((*pspm)->snd_1xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_2xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_3xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_4xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_5xxs));
  if (i != 0)
    goto psi_error5;
  i = sfp_snd_func_tab_init (&((*pspm)->snd_6xxs));
  if (i != 0)
    goto psi_error5;

  return 0;

psi_error5:
  module_release ((*pspm)->module);
  pspm_sfp_free (*pspm);
  sfree (*pspm);
  return -1;
psi_error4:
  fifo_free ((*pspm)->sfull_request);
  sfree ((*pspm)->sfull_request);
psi_error3:
  smutex_destroy ((*pspm)->mut_sfp_pending);
psi_error2:
  module_release ((*pspm)->module);
  module_free ((*pspm)->module);
  sfree ((*pspm)->module);
psi_error1:
  sfree (*pspm);
  return -1;
}

void
pspm_sfp_free (pspm_sfp_t * pspm)
{
  sfp_plugin_t *p;
  psp_req_t *req;

  if (pspm == NULL)
    return;

  sfp_fwd_func_tab_free (pspm->fwd_invites);
  sfp_fwd_func_tab_free (pspm->fwd_acks);
  sfp_fwd_func_tab_free (pspm->fwd_registers);
  sfp_fwd_func_tab_free (pspm->fwd_byes);
  sfp_fwd_func_tab_free (pspm->fwd_optionss);
  sfp_fwd_func_tab_free (pspm->fwd_infos);
  sfp_fwd_func_tab_free (pspm->fwd_cancels);
  sfp_fwd_func_tab_free (pspm->fwd_notifys);
  sfp_fwd_func_tab_free (pspm->fwd_subscribes);
  sfp_fwd_func_tab_free (pspm->fwd_unknowns);

  sfp_rcv_func_tab_free (pspm->rcv_1xxs);
  sfp_rcv_func_tab_free (pspm->rcv_2xxs);
  sfp_rcv_func_tab_free (pspm->rcv_3xxs);
  sfp_rcv_func_tab_free (pspm->rcv_4xxs);
  sfp_rcv_func_tab_free (pspm->rcv_5xxs);
  sfp_rcv_func_tab_free (pspm->rcv_6xxs);

  sfp_snd_func_tab_free (pspm->snd_1xxs);
  sfp_snd_func_tab_free (pspm->snd_2xxs);
  sfp_snd_func_tab_free (pspm->snd_3xxs);
  sfp_snd_func_tab_free (pspm->snd_4xxs);
  sfp_snd_func_tab_free (pspm->snd_5xxs);
  sfp_snd_func_tab_free (pspm->snd_6xxs);


  sfree (pspm->fwd_invites);
  sfree (pspm->fwd_acks);
  sfree (pspm->fwd_registers);
  sfree (pspm->fwd_byes);
  sfree (pspm->fwd_optionss);
  sfree (pspm->fwd_infos);
  sfree (pspm->fwd_cancels);
  sfree (pspm->fwd_notifys);
  sfree (pspm->fwd_subscribes);
  sfree (pspm->fwd_unknowns);

  sfree (pspm->rcv_1xxs);
  sfree (pspm->rcv_2xxs);
  sfree (pspm->rcv_3xxs);
  sfree (pspm->rcv_4xxs);
  sfree (pspm->rcv_5xxs);
  sfree (pspm->rcv_6xxs);

  sfree (pspm->snd_1xxs);
  sfree (pspm->snd_2xxs);
  sfree (pspm->snd_3xxs);
  sfree (pspm->snd_4xxs);
  sfree (pspm->snd_5xxs);
  sfree (pspm->snd_6xxs);

  module_free (pspm->module);
  sfree (pspm->module);

  req = fifo_tryget (pspm->sfull_request);
  while (req != NULL)
    {
      psp_req_free (req);
      sfree (req);
      req = fifo_tryget (pspm->sfull_request);
    }

  fifo_free (pspm->sfull_request);
  sfree (pspm->sfull_request);

  req = fifo_tryget (pspm->sfull_response);
  while (req != NULL)
    {
      /* psp_req_free(req);
         sfree(req);  those request may exist and be deleted somewhere else */
      req = fifo_tryget (pspm->sfull_response);
    }

  fifo_free (pspm->sfull_response);
  sfree (pspm->sfull_response);

  smutex_destroy (pspm->mut_sfp_pending);
  sfree (pspm->mut_sfp_pending);

  for (req = pspm->sfp_pendings; req != NULL; req = pspm->sfp_pendings)
    {
      REMOVE_ELEMENT (pspm->sfp_pendings, req);
      psp_req_free (req);
      sfree (req);
    }

  for (req = pspm->req_terminated; req != NULL; req = pspm->req_terminated)
    {
      REMOVE_ELEMENT (pspm->req_terminated, req);
      psp_req_free (req);
      sfree (req);
    }

  for (p = pspm->sfp_plugins; p != NULL; p = pspm->sfp_plugins)
    {
      REMOVE_ELEMENT (pspm->sfp_plugins, p);
      sfp_plugin_free (p);
      sfree (p);
    }

}


int
pspm_sfp_release (pspm_sfp_t * pspm)
{
  int i;

  i = module_release (pspm->module);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_sfp_start (pspm_sfp_t * pspm, void *(*func_start) (void *), void *arg)
{
  int i;

  i = module_start (pspm->module, func_start, arg);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_sfp_lock_sfp_pending_access (pspm_sfp_t * pspm)
{
  return smutex_lock (pspm->mut_sfp_pending);
}

int
pspm_sfp_unlock_sfp_pending_access (pspm_sfp_t * pspm)
{
  return smutex_unlock (pspm->mut_sfp_pending);
}

int
pspm_sfp_execute (pspm_sfp_t * pspm, int sec_max, int usec_max, int max)
{
  psp_req_t *req;
  fd_set sfp_fdset;
  int max_fd;
  struct timeval tv;

  if (pspm == NULL)
    return -1;

  while (1)
    {
      int s;
      int i;

      tv.tv_sec = sec_max;
      tv.tv_usec = usec_max;

      max--;
      FD_ZERO (&sfp_fdset);
      s = ppl_pipe_get_read_descr (pspm->module->wakeup);
      if (s <= 1)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "No wakeup value in module sfp!\n"));
          return -1;
        }
      FD_SET (s, &sfp_fdset);

      max_fd = s;

      if ((sec_max == -1) || (usec_max == -1))
        i = select (max_fd + 1, &sfp_fdset, NULL, NULL, NULL);
      else
        i = select (max_fd + 1, &sfp_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &sfp_fdset))
        {                       /* time to wake up: we receive "q" to quit and any other char to read all socket */
          char tmp[2];

          i = ppl_pipe_read (pspm->module->wakeup, tmp, 1);
          if (i == 1 && tmp[0] == 'q')
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_INFO2, NULL,
                           "Exiting sfp module!\n"));
              return 0;
            }
        }

      pspm_sfp_lock_sfp_pending_access (pspm);

      /* req_t element containing a response coming from imp */
      req = fifo_tryget (pspm->sfull_response);
      while (req != NULL)
        {
          pspm_sfp_handle_psp_response (pspm, req);
          req = fifo_tryget (pspm->sfull_response);
        }

      pspm_sfp_handle_new_location (pspm);

      /* psp_req element containing a request coming from imp */
      req = fifo_tryget (pspm->sfull_request);
      while (req != NULL)
        {
          pspm_sfp_handle_psp_request (pspm, req);
          req = fifo_tryget (pspm->sfull_request);
        }


      pspm_sfp_unlock_sfp_pending_access (pspm);

      if (max == 0)
        return 0;
    }
}

int
pspm_sfp_create_sfp_branch (pspm_sfp_t * pspm, psp_req_t * req, url_t * url,
                            ppl_dns_ip_t * ips)
{
  transaction_t *transaction;
  sipevent_t *evt;
  osip_ctx_t *osip_ctx;
  sfp_branch_t *br;
  unsigned long int one_inet_addr;
  int i;
  url_t *url2;

  /* create a sfull context if this is the first branch */
  if (req->rsp_ctx == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO4, NULL,
                   "sfp module: The initial transaction for a transaction will be created!\n"));
      i = psp_core_sfp_new_rsp_ctx (&(req->rsp_ctx), req);
      if (i != 0)
        {
          req->rsp_ctx = NULL;
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "sfp module: failed to create a rsp_ctx for this incoming transaction!\n"));
          return -1;
        }
    }

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "sfp module: Creating new branch for statefull transaction!\n"));
  i = sfp_branch_init (&br, req->rsp_ctx->branch);
  if (i != 0)
    return -1;

  i = url_clone (url, &url2);
  if (i != 0)
    {
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  /* if an ips->name is not an IP: we replace the url host with the FQDN.
     (In this case, we are sure ips->name is not a domain name but a real FQDN.) */
  if (ips != NULL && (int) (one_inet_addr = inet_addr (ips->name)) == -1)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "sfp module: Replacing with the FQDN (%s) for statefull transaction!\n",
                   ips->name));
      sfree (url2->host);
      url2->host = sgetcopy (ips->name);
    }

  sfp_branch_set_url (br, url2);

  if (ips != NULL && ips->next != NULL) /* fallback exists */
    sfp_branch_clone_and_set_fallbacks (br, ips->next);

  i = url_clone (br->url, &url2);
  if (i != 0)
    {
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  /* clone the request... start the new transaction... */
  i = osip_msg_sfp_build_request_to_forward (req, &(br->request), url2,
                                             req->request,
                                             __IS_PSP_STAY_ON_PATH (req->flag));
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: Could not create the request to forward for statefull transaction!\n"));
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }
  i = pspm_sfp_call_plugins_for_request (pspm, req, br);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: Could not call plugins for this request!\n"));
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  if (!MSG_IS_INVITE (br->request))
    /* I set the timeout to a lower value. I think it should be 160 for INVITE
       and 32 for other requests. I prefer to use lower value by now, but those
       values will be dynamic later... */
    sfp_branch_set_timeout (br, 32);

  /* create a new outgoing transaction and add the REQUEST */
  evt = osip_new_outgoing_sipmessage (br->request);
  br->request = NULL;           /* avoid double free and speed the proxy... */

  transaction = psp_core_uap_create_transaction (evt);
  if (transaction == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: Could not create a transaction for this request!\n"));
      sfp_branch_free (br);
      sfree (br);
      return -1;
    }

  /* The oSIP stack transaction element has an entry for the resolved IP
     address where the stack should send the request. As we are using
     probably using a FQDN in the rq-uri (or the route), we can force
     the stack to keep the FQDN value there while it still can access
     the real IP of this host. */
  if (ips != NULL)
    {
      /* ! The ips list contains 5060 as a default port for all resolution
         except for domain resoltions.
         So, I'll trust the port only if none is proposed in the rquri or
         route header. This *quick hack* won't work in all situations...
         TO BE FIXED. */
      if (ntohs (ips->sin.sin_port) != 5060)    /* we keep it */
        {
      } else
        {
          if (br->url->port != NULL)
            ips->sin.sin_port = htons ((short) satoi (br->url->port));
        }

      if (MSG_IS_INVITE (req->request))
        ict_set_destination (transaction->ict_context,
                             sgetcopy (inet_ntoa (ips->sin.sin_addr)),
                             ntohs (ips->sin.sin_port));
      else
        nict_set_destination (transaction->nict_context,
                              sgetcopy (inet_ntoa (ips->sin.sin_addr)),
                              ntohs (ips->sin.sin_port));
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "sfp module: The destination for the transaction is set to: '%s'.!\n",
                   inet_ntoa (ips->sin.sin_addr)));
    }
  /* else, the usual processing is done */
  transaction_add_event (transaction, evt);
  evt->transactionid = transaction->transactionid;
  sfp_branch_set_osip_id (br, transaction->transactionid);

  osip_ctx_init (&osip_ctx, transaction);

  ADD_ELEMENT (req->rsp_ctx->branches, br);
  osip_ctx_start (osip_ctx);

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO2, NULL,
               "sfp module: transaction correctly forwarded!\n"));
  return 0;
}

int
pspm_sfp_handle_ack (pspm_sfp_t * pspm, psp_req_t * req)
{
  location_t *ptr;
  location_t *loc;
  int i;

  /* For statefull mode, the plugins have to provide
     locations to be resolved... */
  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      int is_an_ip = inet_addr (loc->url->host);

      ptr = loc->next;
      if (is_an_ip == -1)       /* not an ip address, ask for a resolution */
        {
          psp_core_event_resolv_add_query (sgetcopy (loc->url->host));
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "sfp module: we have to send the ACK now, but the ACK contains a FQDN!\n"));
      } else
        {
          /* WE DON'T NEED TO WAIT! this location can be tried right now. */
          sip_t *ack;
          int port;

          i = osip_msg_sfp_build_request_to_forward (req, &ack, loc->url,
                                                     req->request,
                                                     __IS_PSP_STAY_ON_PATH (req->
                                                                            flag));
          if (i != 0)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "sfp module: could not build ACK to forward!\n"));
          } else
            {
              i = pspm_sfp_call_plugins_for_request (pspm, req, NULL);
              if (i != 0)
                {
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_ERROR, NULL,
                               "sfp module: Could not call plugins for this request!\n"));
                }
            }
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "sfp module: we have to send the ACK now!\n"));

          port = 5060;
          if (loc->url->port != NULL)
            port = satoi (loc->url->port);

          psp_core_cb_snd_message (NULL, ack, loc->url->host, port, -1);
          msg_free (ack);
          sfree (ack);
          loc->url = NULL;

          REMOVE_ELEMENT (req->locations, loc);
          location_free (loc);
          sfree (loc);
        }
    }
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_ERROR, NULL,
               "sfp module: deleting ACK!\n"));
  psp_req_free (req);
  sfree (req);
  return 0;
}

int
pspm_sfp_handle_psp_request (pspm_sfp_t * pspm, psp_req_t * req)
{
  transaction_t *transaction;
  sipevent_t *evt;
  osip_ctx_t *osip_ctx;
  sip_t *answer1xx;
  location_t *loc;
  location_t *ptr;
  int i;

  if (MSG_IS_ACK (req->request))
    return pspm_sfp_handle_ack (pspm, req);

  /* when a NEW request is to be handled in statefull mode, the sfp module
     has to create an incoming transasction as soon as possible. */
  evt = osip_new_incoming_sipmessage (req->request);
  evt->transactionid = 0;
  req->request = NULL;          /* avoid duplicate if not needed */
  /* just to be sure this request is a new one (we may receive
     a retransmission before we create the transaction, so the
     retransmission was not detected at first) */
  i = psp_core_find_transaction_and_add_event (evt);
  if (i == 0)                   /* detect a retransmission... just delete the
                                   psp_req element */
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_WARNING, NULL,
                   "sfp module: this was a retransmission.!\n"));
      psp_req_free (req);
      sfree (req);
      return 0;                 /* ok */
    }

  i = msg_clone (evt->sip, &(req->request));
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: could not clone request. (discard it)!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      psp_req_free (req);
      sfree (req);
      return -1;
    }

  transaction = psp_core_uap_create_transaction (evt);
  if (transaction == NULL)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: could not create a transaction for this request. (discard it)!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      psp_req_free (req);
      sfree (req);
      return -1;
    }
  transaction_add_event (transaction, evt);
  psp_req_set_osip_id (req, transaction->transactionid);
  osip_ctx_init (&osip_ctx, transaction);
  osip_ctx_start (osip_ctx);
  /* now, all retransmissions will be handled by oSIP. */

  i = osip_msg_sfp_build_response (&answer1xx, 100, req->request);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "sfp module: could not create a 100 Trying for this transaction. (discard it and let the transaction die itself)!\n"));
      psp_req_free (req);
      sfree (req);
      return -1;
    }

  /* pspm_sfp_lock_sfp_pending_access(pspm); */
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "sfp module: Request added in sfp_pendings!\n"));
  ADD_ELEMENT (pspm->sfp_pendings, req);

  evt = osip_new_outgoing_sipmessage (answer1xx);
  evt->transactionid = transaction->transactionid;
  transaction_add_event (transaction, evt);

  /* For statefull mode, the plugins have to provide
     locations to be resolved... */
  for (loc = req->locations; loc != NULL; loc = ptr)
    {
      int is_an_ip = inet_addr (loc->url->host);

      ptr = loc->next;
      if (is_an_ip == -1)       /* not an ip address, ask for a resolution */
        {
          psp_core_event_resolv_add_query (sgetcopy (loc->url->host));
      } else
        {
          /* WE DON'T NEED TO WAIT! this location can be tried right now. */
          i = pspm_sfp_create_sfp_branch (pspm, req, loc->url, NULL);
          /* in all cases, (failure or success), this location
             has been tried and will never be tried again */
          REMOVE_ELEMENT (req->locations, loc);
          location_free (loc);
          sfree (loc);
        }
    }
  /* pspm_sfp_unlock_sfp_pending_access(pspm); */

  return 0;
}

int
pspm_sfp_handle_new_location (pspm_sfp_t * pspm)
{
  location_t *loc;
  location_t *ptr;
  psp_req_t *req_ptr;
  psp_req_t *req;
  int unfinished;
  int i;

  /* this method is regularly called and check if some pending
     statefull transactions have new locations to try. */

  /*  pspm_sfp_lock_sfp_pending_access(pspm); */
  for (req = pspm->sfp_pendings; req != NULL; req = req_ptr)
    {
      req_ptr = req->next;
      for (loc = req->locations; loc != NULL; loc = ptr)
        {
          int is_an_ip = inet_addr (loc->url->host);

          ptr = loc->next;
          if (is_an_ip == -1)   /* not an ip address, ask for a resolution */
            {
              ppl_dns_entry_t *dns_result;

              i = ppl_dns_get_result (&dns_result, loc->url->host);
              if (i != 0)
                {
                  /* No available IP for this host */
                  /* How do we make a difference between failed resolution
                     and ones that are pendings... TODO */
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_INFO3, NULL,
                               "module sfp: Keep waiting for the resolution to succeed...!\n"));

              } else
                {
                  /* The resolver can provide several IPs for the same location.
                     This is a major issue here: What should I do: starting a
                     transactions for each IP does not seems to be reasonnable.
                     (forking should be done on several locations, not
                     on one location that resolved to many IPs. In the case
                     where a host is resolved to many IPs, I sould try the
                     first one and use the other ones as fallbacks.
                   */
                  i = pspm_sfp_create_sfp_branch (pspm, req,
                                                  loc->url, dns_result->dns_ips);
                  /* in all cases, (failure or success), this location
                     has been tried and will never be tried again */
                  REMOVE_ELEMENT (req->locations, loc);
                  location_free (loc);
                  sfree (loc);
                }
          } else
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_BUG, NULL,
                           "module sfp: An ip address is found here! (I don't expect it here)!\n"));
            }
        }
      unfinished = 0;
      for (loc = req->locations; loc != NULL; loc = ptr)
        {
          ppl_dns_error_t *error;

          ptr = loc->next;
          if (0 == ppl_dns_get_error (&error, loc->url->host))
            {
              REMOVE_ELEMENT (req->locations, loc);
              location_free (loc);
              sfree (loc);
          } else
            unfinished = 1;     /* there is hope! */
        }

      if (unfinished == 0
          && (req->rsp_ctx != NULL && req->rsp_ctx->branches != NULL))
        {                       /* if all locations are consumed */
          sfp_branch_t *br;
          sfp_branch_t *ptr;

          /* all transactions have no answer since a long time, we delete it. */
          for (br = req->rsp_ctx->branches; br != NULL; br = ptr)
            {
              ptr = br->next;
              if (br->response == NULL && 0 == sfp_branch_is_expired (br))      /* no answer yet! */
                {               /* expired branch with no informational and final answer... */
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_INFO1, NULL,
                               "module sfp: A branch has expired without being answered (no informationnal answer)?!\n"));
                  /* we must NOT send any CANCEL (until a informationnal answer
                     is received. */
                  REMOVE_ELEMENT (req->rsp_ctx->branches, br);
                  sfp_branch_free (br);
                  sfree (br);
              } else if (br->response != NULL && MSG_IS_STATUS_1XX (br->response)
                         && 0 == sfp_branch_is_expired (br))
                {               /* expired branch with no final answer... */
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_INFO1, NULL,
                               "module sfp: A branch has expired without being answered (But a informationnal answer has been received)!\n"));
                  /* we MUST send a CANCEL... TODO */
                  REMOVE_ELEMENT (req->rsp_ctx->branches, br);
                  sfp_branch_free (br);
                  sfree (br);
                }
            }
        }

      if (0 == psp_req_is_expired (req))
        {
          sfp_branch_t *br;
          sfp_branch_t *ptr = NULL;

          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                                  "slp module: Request has expired! Searching for a final answer!\n"));
          if (req->rsp_ctx != NULL && req->rsp_ctx->branches != NULL)
            for (br = req->rsp_ctx->branches; br != NULL; br = br->next)
              {
                if (br->response != NULL && MSG_IS_STATUS_2XX (br->response))
                  {
                    ptr = br;
                    break;
                } else if (br->response != NULL
                           && !MSG_IS_STATUS_1XX (br->response))
                  {
                    if (ptr == NULL)
                      ptr = br;
                    else if (0 < strcmp (ptr->response->strtline->statuscode,
                                         br->response->strtline->statuscode))
                      ptr = br; /* better response */
                  }
              }
          if (ptr != NULL)      /* one final answer has been received. We forward it */
            {
              sip_t *responsexxx;

              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_INFO3, NULL,
                           "module sfp: We have a xxx>299 response, let's try to forward it!\n"));
              /* we have a 2xx final answer! forward it... */
              i =
                osip_msg_sfp_build_response_to_forward (&responsexxx,
                                                        ptr->response);
              if (i != 0)
                {
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_ERROR, NULL,
                               "module sfp: Cannot forward xxx response!\n"));
                  REMOVE_ELEMENT (pspm->sfp_pendings, req);
                  psp_req_free (req);
                  sfree (req);
              } else
                {
                  /* the new xxx is the "best answer" */
                  if (req->response != NULL)
                    {
                      msg_free (req->response);
                      sfree (req->response);
                      req->response = NULL;
                    }
                  i = msg_clone (responsexxx, &(req->response));
                  if (i != 0)
                    {
                      OSIP_TRACE (osip_trace
                                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                   "module sfp: Cannot make a copy of forwarded response!\n"));
                      msg_free (responsexxx);
                      sfree (responsexxx);
                      REMOVE_ELEMENT (pspm->sfp_pendings, req);
                      psp_req_free (req);
                      sfree (req);
                  } else
                    {
                      i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
                      if (i != 0)
                        {
                          OSIP_TRACE (osip_trace
                                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                       "module sfp: Could not call plugins!\n"));
                          msg_free (responsexxx);
                          sfree (responsexxx);
                          REMOVE_ELEMENT (pspm->sfp_pendings, req);
                          psp_req_free (req);
                          sfree (req);
                      } else
                        {
                          sipevent_t *evt;

                          evt = osip_new_outgoing_sipmessage (responsexxx);
                          evt->transactionid = psp_req_get_osip_id (req);
                          i = psp_core_find_transaction_and_add_event (evt);
                          if (i != 0)
                            {
                              OSIP_TRACE (osip_trace
                                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                           "module sfp: Cannot find the transaction for the event!\n"));
                              msg_free (evt->sip);
                              sfree (evt->sip);
                              sfree (evt);
                              REMOVE_ELEMENT (pspm->sfp_pendings, req);
                              psp_req_free (req);
                              sfree (req);
                          } else
                            {
                              OSIP_TRACE (osip_trace
                                          (__FILE__, __LINE__, OSIP_INFO3, NULL,
                                           "module sfp: Response 2xx forwarded!\n"));
                              /* if we remove the psp_req element here, the ACK will not
                                 be related to this transaction. (I don't mind, but this
                                 may be fixed one day.) */
                              REMOVE_ELEMENT (pspm->sfp_pendings, req);
                              psp_req_free (req);
                              sfree (req);
                            }
                        }
                    }
                }
              /* One final answer has been forwarded, or at least we failed! */
              /* Anyway, we are sure that req ressource are released: */
              req = NULL;
            }
        }
      if (req == NULL)
        {
      } else if (unfinished == 0
                 && (req->rsp_ctx == NULL || req->rsp_ctx->branches == NULL))
        {
          /* We have definitive information about this request! No
             locations have been found for this transaction. */
          sipevent_t *evt;
          sip_t *response480;

          /* return 480 Not available? */
          if (req->response != NULL)
            {
              msg_free (req->response);
              sfree (req->response);
              req->response = NULL;
            }
          i = osip_msg_sfp_build_response (&response480, 480, req->request);
          if (i != 0)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "module sfp: Cannot build 480 response!\n"));
              REMOVE_ELEMENT (pspm->sfp_pendings, req);
              psp_req_free (req);
              sfree (req);
          } else
            {
              evt = osip_new_outgoing_sipmessage (response480);
              evt->transactionid = psp_req_get_osip_id (req);
              i = psp_core_find_transaction_and_add_event (evt);
              if (i != 0)
                {
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_ERROR, NULL,
                               "module sfp: Cannot find the transaction for the event!\n"));
                  msg_free (evt->sip);
                  sfree (evt->sip);
                  sfree (evt);
                  REMOVE_ELEMENT (pspm->sfp_pendings, req);
                  psp_req_free (req);
                  sfree (req);
              } else
                {
                  i = msg_clone (response480, &req->response);
                  if (i != 0)
                    {
                      OSIP_TRACE (osip_trace
                                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                   "module sfp: Cannot build a copy of the 480 answer!\n"));
                      msg_free (evt->sip);
                      sfree (evt->sip);
                      sfree (evt);
                      REMOVE_ELEMENT (pspm->sfp_pendings, req);
                      psp_req_free (req);
                      sfree (req);
                  } else
                    {
                      OSIP_TRACE (osip_trace
                                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                   "module sfp: No valid location found; answer to request with 480.!\n"));
                      /* Definitive answer has to be sent:
                         Call plugins: */
                      i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
                      if (i != 0)
                        {
                          OSIP_TRACE (osip_trace
                                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                                       "module sfp: Can't call plugins for response.!\n"));
                          msg_free (evt->sip);
                          sfree (evt->sip);
                          sfree (evt);
                          REMOVE_ELEMENT (pspm->sfp_pendings, req);
                          psp_req_free (req);
                          sfree (req);
                        }
                      /* if we remove the psp_req element here, the ACK will not
                         be related to this transaction. (I don't mind, but this
                         may be fixed one day.) */
                      REMOVE_ELEMENT (pspm->sfp_pendings, req);
                      psp_req_free (req);
                      sfree (req);
                    }
                }
              /* This code has to be cleaned...   :) */
            }
        }
    }
  return 0;
}


int
pspm_sfp_handle_psp_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sipevent_t *evt;
  sip_t *responsexxx;
  sfp_branch_t *sfp;
  sip_t *best_response;
  int i;

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "module sfp: pspm_sfp_handle_psp_response!\n"));

  /* BUG BUG:
     If the request has expired without being answered, it can be
     released and the request can still exist here... We have no
     way to guess that. Instead, we will differ the release of
     context when they are tagged as used. */

  /* if a 200 OK has been received, we can end this transaction */
  best_response = NULL;
  for (sfp = req->rsp_ctx->branches; sfp != NULL; sfp = sfp->next)
    {
      if (sfp->response != NULL && MSG_IS_STATUS_2XX (sfp->response))
        {
          best_response = sfp->response;
          break;
        }
    }

  if (best_response != NULL)
    {

      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "module sfp: We have a xxx>200 response, let's try to forward it!\n"));
      /* we have a 2xx final answer! forward it... */
      i = osip_msg_sfp_build_response_to_forward (&responsexxx, best_response);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "module sfp: Cannot forward 2xx response!\n"));
          REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
          sfp_branch_free (sfp);
          sfree (sfp);
          return -1;
        }

      /* the new 2xx is the "best answer" */
      if (req->response != NULL)
        {
          msg_free (req->response);
          sfree (req->response);
          req->response = NULL;
        }
      i = msg_clone (responsexxx, &(req->response));
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "module sfp: Cannot make a copy of forwarded response!\n"));
          msg_free (responsexxx);
          sfree (responsexxx);
          REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
          sfp_branch_free (sfp);
          sfree (sfp);
          return -1;
        }

      i = pspm_sfp_call_plugins_for_snd_response (pspm, req);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "module sfp: Could not call plugins!\n"));
          REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
          sfp_branch_free (sfp);
          sfree (sfp);
          msg_free (req->response);
          sfree (req->response);
          req->response = NULL;
          return -1;
        }

      evt = osip_new_outgoing_sipmessage (responsexxx);
      evt->transactionid = psp_req_get_osip_id (req);
      i = psp_core_find_transaction_and_add_event (evt);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "module sfp: Cannot find the transaction for the event!\n"));
          msg_free (evt->sip);
          sfree (evt->sip);
          sfree (evt);
          REMOVE_ELEMENT (req->rsp_ctx->branches, sfp);
          sfp_branch_free (sfp);
          sfree (sfp);
          msg_free (req->response);
          sfree (req->response);
          req->response = NULL;
          return -1;
        }

      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO3, NULL,
                   "module sfp: Response 2xx forwarded!\n"));
      /* if we remove the psp_req element here, the ACK will not
         be related to this transaction. (I don't mind, but this
         may be fixed one day.) */
      REMOVE_ELEMENT (pspm->sfp_pendings, req);
      psp_req_free (req);
      sfree (req);
      return 0;
    }
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "module sfp: Waiting for final answer!\n"));

  return 0;
}

int
pspm_sfp_call_plugins_for_request (pspm_sfp_t * pspm, psp_req_t * req,
                                   sfp_branch_t * sfp_branch)
{
  sfp_fwd_func_tab_t *t;
  sfp_fwd_func_t *f;

  if (MSG_IS_INVITE (req->request))
    t = pspm->fwd_invites;
  else if (MSG_IS_ACK (req->request))
    t = pspm->fwd_acks;
  else if (MSG_IS_REGISTER (req->request))
    t = pspm->fwd_registers;
  else if (MSG_IS_BYE (req->request))
    t = pspm->fwd_byes;
  else if (MSG_IS_OPTIONS (req->request))
    t = pspm->fwd_optionss;
  else if (MSG_IS_INFO (req->request))
    t = pspm->fwd_infos;
  else if (MSG_IS_CANCEL (req->request))
    t = pspm->fwd_cancels;
  else if (MSG_IS_NOTIFY (req->request))
    t = pspm->fwd_notifys;
  else if (MSG_IS_SUBSCRIBE (req->request))
    t = pspm->fwd_subscribes;
  else
    t = pspm->fwd_unknowns;

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}

int
pspm_sfp_call_plugins_for_rcv_response (pspm_sfp_t * pspm, psp_req_t * req,
                                        sfp_branch_t * branch)
{
  sfp_rcv_func_tab_t *t;
  sfp_rcv_func_t *f;

  if (MSG_IS_STATUS_1XX (branch->response))
    t = pspm->rcv_1xxs;
  else if (MSG_IS_STATUS_2XX (branch->response))
    t = pspm->rcv_2xxs;
  else if (MSG_IS_STATUS_3XX (branch->response))
    t = pspm->rcv_3xxs;
  else if (MSG_IS_STATUS_4XX (branch->response))
    t = pspm->rcv_4xxs;
  else if (MSG_IS_STATUS_5XX (branch->response))
    t = pspm->rcv_5xxs;
  else if (MSG_IS_STATUS_6XX (branch->response))
    t = pspm->rcv_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}


int
pspm_sfp_call_plugins_for_snd_response (pspm_sfp_t * pspm, psp_req_t * req)
{
  sfp_snd_func_tab_t *t;
  sfp_snd_func_t *f;

  if (MSG_IS_STATUS_1XX (req->response))
    t = pspm->snd_1xxs;
  else if (MSG_IS_STATUS_2XX (req->response))
    t = pspm->snd_2xxs;
  else if (MSG_IS_STATUS_3XX (req->response))
    t = pspm->snd_3xxs;
  else if (MSG_IS_STATUS_4XX (req->response))
    t = pspm->snd_4xxs;
  else if (MSG_IS_STATUS_5XX (req->response))
    t = pspm->snd_5xxs;
  else if (MSG_IS_STATUS_6XX (req->response))
    t = pspm->snd_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}
