/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <psp_req.h>


int
sfp_branch_init (sfp_branch_t ** sfp_branch, char *branch)
{
  *sfp_branch = (sfp_branch_t *) smalloc (sizeof (sfp_branch_t));
  if (*sfp_branch == NULL)
    return -1;
  (*sfp_branch)->branch = NULL;
  (*sfp_branch)->osip_id = -1;
  (*sfp_branch)->ctx_start = time (NULL);
  (*sfp_branch)->ctx_timeout = 15;      /* this is the default delay? */
  (*sfp_branch)->url = NULL;
  (*sfp_branch)->fallbacks = NULL;
  (*sfp_branch)->response = NULL;
  (*sfp_branch)->next = NULL;
  (*sfp_branch)->parent = NULL;
  return 0;
}

void
sfp_branch_free (sfp_branch_t * branch)
{
  ppl_dns_ip_t *f;

  if (branch == NULL)
    return;

  sfree (branch->branch);
  url_free (branch->url);
  sfree (branch->url);

  for (f = branch->fallbacks; f != NULL; f = f->next)
    {
      REMOVE_ELEMENT (branch->fallbacks, f);
      ppl_dns_ip_free (f);
      sfree (f);
    }

  msg_free (branch->request);
  sfree (branch->request);
  msg_free (branch->response);
  sfree (branch->response);
}

int
sfp_branch_get_osip_id (sfp_branch_t * branch)
{
  if (branch == NULL)
    return -1;
  return branch->osip_id;
}

int
sfp_branch_set_osip_id (sfp_branch_t * branch, int osip_id)
{
  if (branch == NULL)
    return -1;
  branch->osip_id = osip_id;
  return 0;
}

int
sfp_branch_set_url (sfp_branch_t * branch, url_t * url)
{
  if (branch == NULL)
    return -1;
  branch->url = url;
  return 0;
}

int
sfp_branch_set_timeout (sfp_branch_t * branch, int delay)
{
  if (branch == NULL)
    return -1;
  branch->ctx_timeout = delay;
  return 0;
}

int
sfp_branch_is_expired (sfp_branch_t * branch)
{
  ppl_time_t now;

  if (branch == NULL)
    return -1;
  now = ppl_time ();
  if (now - branch->ctx_start >= branch->ctx_timeout)
    return 0;
  return -1;
}

int
sfp_branch_clone_and_set_fallbacks (sfp_branch_t * branch, ppl_dns_ip_t * ips)
{
  ppl_dns_ip_t *f;
  ppl_dns_ip_t *ip;
  int i;

  if (branch == NULL)
    return -1;
  for (ip = ips; ip != NULL; ip = ip->next)
    {
      i = ppl_dns_ip_clone (ip, &f);
      if (i == 0)
        {
          ADD_ELEMENT (branch->fallbacks, f);
        }
    }
  return 0;
}

int
sfp_branch_set_request (sfp_branch_t * branch, sip_t * req)
{
  if (branch == NULL)
    return -1;
  branch->request = req;
  return 0;
}

int
sfp_branch_set_response (sfp_branch_t * branch, sip_t * resp)
{
  if (branch == NULL)
    return -1;
  branch->response = resp;
  return 0;
}

/* useless */
/*
  int  sfp_branch_match(sfp_branch_t *branch, char *branch2)
  {
  if (branch==NULL || branch2==NULL) return -1;
  if (strlen(branch2)!=43) return -1; //BUG? branch MUST be of size '43'
  if (0==strncmp(branch->branch, branch2, 43))
  {
  return 0;
  }
  return -1;
  }
*/
int
rsp_ctx_init (rsp_ctx_t ** rsp_ctx, char *branch)
{
  (*rsp_ctx) = (rsp_ctx_t *) smalloc (sizeof (rsp_ctx_t));
  if (*rsp_ctx == NULL)
    return -1;

  (*rsp_ctx)->fork = 0;
  (*rsp_ctx)->ind = 0;
  (*rsp_ctx)->branch = branch;
  (*rsp_ctx)->branches = NULL;
  return 0;
}

void
rsp_ctx_free (rsp_ctx_t * rsp_ctx)
{
  sfp_branch_t *b;

  if (rsp_ctx == NULL)
    return;
  for (b = rsp_ctx->branches; b != NULL; b = rsp_ctx->branches)
    {
      rsp_ctx->branches = b->next;
      sfp_branch_free (b);
      sfree (b);
    }
  sfree (rsp_ctx->branch);
}

int
rsp_ctx_set_fork (rsp_ctx_t * rsp_ctx)
{
  if (rsp_ctx == NULL)
    return -1;
  rsp_ctx->fork = 1;            /* TRUE */
  return 0;
}

int
rsp_ctx_set_sequential (rsp_ctx_t * rsp_ctx)
{
  if (rsp_ctx == NULL)
    return -1;
  rsp_ctx->fork = 0;            /* FALSE */
  return 0;
}
