/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>
#include <osip_msg.h>

int
pspm_slp_init (pspm_slp_t ** pspm)
{
  int i;

  (*pspm) = (pspm_slp_t *) smalloc (sizeof (pspm_slp_t));
  if (*pspm == NULL)
    return -1;

  i = module_init (&((*pspm)->module), MOD_THREAD, "slp");
  if (i != 0)
    goto psi_error1;
  (*pspm)->slp_plugins = NULL;


  (*pspm)->mut_slp_pending = smutex_init ();
  if ((*pspm)->mut_slp_pending == NULL)
    goto psi_error2;
  (*pspm)->slp_pending = NULL;

  (*pspm)->sless_traffic = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sless_traffic == NULL)
    goto psi_error3;
  fifo_init ((*pspm)->sless_traffic);

  (*pspm)->sless_request = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sless_request == NULL)
    goto psi_error4;
  fifo_init ((*pspm)->sless_request);

  (*pspm)->sless_response = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->sless_response == NULL)
    goto psi_error5;
  fifo_init ((*pspm)->sless_response);

  (*pspm)->fwd_invites = NULL;
  (*pspm)->fwd_acks = NULL;
  (*pspm)->fwd_registers = NULL;
  (*pspm)->fwd_byes = NULL;
  (*pspm)->fwd_optionss = NULL;
  (*pspm)->fwd_infos = NULL;
  (*pspm)->fwd_cancels = NULL;
  (*pspm)->fwd_notifys = NULL;
  (*pspm)->fwd_subscribes = NULL;
  (*pspm)->fwd_unknowns = NULL;

  (*pspm)->rcv_1xxs = NULL;
  (*pspm)->rcv_2xxs = NULL;
  (*pspm)->rcv_3xxs = NULL;
  (*pspm)->rcv_4xxs = NULL;
  (*pspm)->rcv_5xxs = NULL;
  (*pspm)->rcv_6xxs = NULL;

  (*pspm)->snd_1xxs = NULL;
  (*pspm)->snd_2xxs = NULL;
  (*pspm)->snd_3xxs = NULL;
  (*pspm)->snd_4xxs = NULL;
  (*pspm)->snd_5xxs = NULL;
  (*pspm)->snd_6xxs = NULL;

  i = slp_fwd_func_tab_init (&((*pspm)->fwd_invites));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_acks));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_registers));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_byes));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_optionss));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_infos));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_cancels));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_notifys));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_subscribes));
  if (i != 0)
    goto psi_error6;
  i = slp_fwd_func_tab_init (&((*pspm)->fwd_unknowns));
  if (i != 0)
    goto psi_error6;

  i = slp_rcv_func_tab_init (&((*pspm)->rcv_1xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_rcv_func_tab_init (&((*pspm)->rcv_2xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_rcv_func_tab_init (&((*pspm)->rcv_3xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_rcv_func_tab_init (&((*pspm)->rcv_4xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_rcv_func_tab_init (&((*pspm)->rcv_5xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_rcv_func_tab_init (&((*pspm)->rcv_6xxs));
  if (i != 0)
    goto psi_error6;

  i = slp_snd_func_tab_init (&((*pspm)->snd_1xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_snd_func_tab_init (&((*pspm)->snd_2xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_snd_func_tab_init (&((*pspm)->snd_3xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_snd_func_tab_init (&((*pspm)->snd_4xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_snd_func_tab_init (&((*pspm)->snd_5xxs));
  if (i != 0)
    goto psi_error6;
  i = slp_snd_func_tab_init (&((*pspm)->snd_6xxs));
  if (i != 0)
    goto psi_error6;

  return 0;

psi_error6:
  module_release ((*pspm)->module);
  pspm_slp_free (*pspm);
  sfree (*pspm);
  return -1;
psi_error5:
  fifo_free ((*pspm)->sless_request);
  sfree ((*pspm)->sless_request);
psi_error4:
  fifo_free ((*pspm)->sless_traffic);
  sfree ((*pspm)->sless_traffic);
psi_error3:
  smutex_destroy ((*pspm)->mut_slp_pending);
  sfree ((*pspm)->mut_slp_pending);
psi_error2:
  module_release ((*pspm)->module);
  module_free ((*pspm)->module);
  sfree ((*pspm)->module);
psi_error1:
  sfree (*pspm);
  return -1;
}

void
pspm_slp_free (pspm_slp_t * pspm)
{
  slp_plugin_t *p;
  sipevent_t *evt;
  psp_req_t *req;
  sip_t *resp;

  if (pspm == NULL)
    return;

  slp_fwd_func_tab_free (pspm->fwd_invites);
  slp_fwd_func_tab_free (pspm->fwd_acks);
  slp_fwd_func_tab_free (pspm->fwd_registers);
  slp_fwd_func_tab_free (pspm->fwd_byes);
  slp_fwd_func_tab_free (pspm->fwd_optionss);
  slp_fwd_func_tab_free (pspm->fwd_infos);
  slp_fwd_func_tab_free (pspm->fwd_cancels);
  slp_fwd_func_tab_free (pspm->fwd_notifys);
  slp_fwd_func_tab_free (pspm->fwd_subscribes);
  slp_fwd_func_tab_free (pspm->fwd_unknowns);

  slp_rcv_func_tab_free (pspm->rcv_1xxs);
  slp_rcv_func_tab_free (pspm->rcv_2xxs);
  slp_rcv_func_tab_free (pspm->rcv_3xxs);
  slp_rcv_func_tab_free (pspm->rcv_4xxs);
  slp_rcv_func_tab_free (pspm->rcv_5xxs);
  slp_rcv_func_tab_free (pspm->rcv_6xxs);

  slp_snd_func_tab_free (pspm->snd_1xxs);
  slp_snd_func_tab_free (pspm->snd_2xxs);
  slp_snd_func_tab_free (pspm->snd_3xxs);
  slp_snd_func_tab_free (pspm->snd_4xxs);
  slp_snd_func_tab_free (pspm->snd_5xxs);
  slp_snd_func_tab_free (pspm->snd_6xxs);


  sfree (pspm->fwd_invites);
  sfree (pspm->fwd_acks);
  sfree (pspm->fwd_registers);
  sfree (pspm->fwd_byes);
  sfree (pspm->fwd_optionss);
  sfree (pspm->fwd_infos);
  sfree (pspm->fwd_cancels);
  sfree (pspm->fwd_notifys);
  sfree (pspm->fwd_subscribes);
  sfree (pspm->fwd_unknowns);

  sfree (pspm->rcv_1xxs);
  sfree (pspm->rcv_2xxs);
  sfree (pspm->rcv_3xxs);
  sfree (pspm->rcv_4xxs);
  sfree (pspm->rcv_5xxs);
  sfree (pspm->rcv_6xxs);

  sfree (pspm->snd_1xxs);
  sfree (pspm->snd_2xxs);
  sfree (pspm->snd_3xxs);
  sfree (pspm->snd_4xxs);
  sfree (pspm->snd_5xxs);
  sfree (pspm->snd_6xxs);

  module_free (pspm->module);
  sfree (pspm->module);

  evt = fifo_tryget (pspm->sless_traffic);
  while (evt != NULL)
    {
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      evt = fifo_tryget (pspm->sless_traffic);
    }

  fifo_free (pspm->sless_traffic);
  sfree (pspm->sless_traffic);

  req = fifo_tryget (pspm->sless_request);
  while (req != NULL)
    {
      psp_req_free (req);
      sfree (req);
      req = fifo_tryget (pspm->sless_request);
    }

  fifo_free (pspm->sless_request);
  sfree (pspm->sless_request);

  resp = fifo_tryget (pspm->sless_response);
  while (resp != NULL)
    {
      msg_free (resp);
      sfree (resp);
      resp = fifo_tryget (pspm->sless_response);
    }

  fifo_free (pspm->sless_response);
  sfree (pspm->sless_response);

  smutex_destroy (pspm->mut_slp_pending);
  sfree (pspm->mut_slp_pending);

  for (req = pspm->slp_pending; req != NULL; req = pspm->slp_pending)
    {
      REMOVE_ELEMENT (pspm->slp_pending, req);
      psp_req_free (req);
      sfree (req);
    }

  for (p = pspm->slp_plugins; p != NULL; p = pspm->slp_plugins)
    {
      REMOVE_ELEMENT (pspm->slp_plugins, p);
      slp_plugin_free (p);
      sfree (p);
    }

}

int
pspm_slp_release (pspm_slp_t * pspm)
{
  int i;

  i = module_release (pspm->module);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_slp_start (pspm_slp_t * pspm, void *(*func_start) (void *), void *arg)
{
  int i;

  i = module_start (pspm->module, func_start, arg);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_slp_lock_slp_pending_access (pspm_slp_t * pspm)
{
  return smutex_lock (pspm->mut_slp_pending);
}

int
pspm_slp_unlock_slp_pending_access (pspm_slp_t * pspm)
{
  return smutex_unlock (pspm->mut_slp_pending);
}

int
pspm_slp_execute (pspm_slp_t * pspm, int sec_max, int usec_max, int max)
{
  psp_req_t *req;
  sipevent_t *evt;
  sip_t *resp;
  fd_set slp_fdset;
  int max_fd;
  struct timeval tv;

  if (pspm == NULL)
    return -1;

  while (1)
    {
      int s;
      int i;

      tv.tv_sec = sec_max;
      tv.tv_usec = usec_max;

      max--;
      FD_ZERO (&slp_fdset);
      s = ppl_pipe_get_read_descr (pspm->module->wakeup);
      if (s <= 1)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "No wakeup value in module slp!\n"));
          return -1;
        }
      FD_SET (s, &slp_fdset);

      max_fd = s;

      if ((sec_max == -1) || (usec_max == -1))
        i = select (max_fd + 1, &slp_fdset, NULL, NULL, NULL);
      else
        i = select (max_fd + 1, &slp_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &slp_fdset))
        {                       /* time to wake up: we receive "q" to quit and any other char to read all socket */
          char tmp[2];

          i = ppl_pipe_read (pspm->module->wakeup, tmp, 1);
          if (i == 1 && tmp[0] == 'q')
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_INFO2, NULL,
                           "Exiting slp module!\n"));
              return 0;
            }
        }

      /* msp_req element containing a request coming from imp */
      evt = fifo_tryget (pspm->sless_traffic);
      while (evt != NULL)
        {
          pspm_slp_check_if_request_belongs_to_this_module (pspm, evt);
          evt = fifo_tryget (pspm->sless_traffic);
        }

      req = fifo_tryget (pspm->sless_request);
      while (req != NULL)
        {
          pspm_slp_handle_psp_request (pspm, req);
          req = fifo_tryget (pspm->sless_request);
        }

      /* msp_req element containing a response coming from imp */
      resp = fifo_tryget (pspm->sless_response);
      while (resp != NULL)
        {
          pspm_slp_handle_psp_response (pspm, resp);
          resp = fifo_tryget (pspm->sless_response);
        }

      /* regurlarly check for finished transactions */
      for (req = pspm->slp_pending; req != NULL;)
        {
          if (req->sl_ctx == NULL)      /* this MUST never happen */
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_BUG, NULL,
                           "slp plugin: no context for stateless transaction?!\n"));
              req = req->next;
          } else if (0 == psp_req_is_expired (req))
            {
              /* Here we could switch to UAS mode and reply 404 or 480 Not found */
              psp_req_t *tmp_req;

              tmp_req = req->next;
              REMOVE_ELEMENT (pspm->slp_pending, req);
              psp_req_free (req);
              sfree (req);
              req = tmp_req;
          } else
            req = req->next;
        }

      if (max == 0)
        return 0;
    }
}

psp_req_t *
pspm_slp_find_pending_req (pspm_slp_t * pspm, char *branch)
{
  psp_req_t *mr;

  for (mr = pspm->slp_pending; mr != NULL; mr = mr->next)
    {
      /* request in slp_pending MUST have a branch parameter */
      if (mr->sl_ctx->branch != NULL && 0 == slp_branch_match (mr->sl_ctx, branch))
        {
          return mr;
        }
    }
  return NULL;
}

int
pspm_slp_forward_response (sip_t * response)
{
  int port;
  via_t *via;

  via = list_get (response->vias, 0);
  if (via == NULL)
    return -1;
  if (via->port == NULL)
    port = 5060;
  else
    port = atoi (via->port);
  return psp_core_cb_snd_message (NULL, response, via->host, port, -1);
}

int
pspm_slp_check_if_request_belongs_to_this_module (pspm_slp_t * pspm,
                                                  sipevent_t * evt)
{
  psp_req_t *req;
  char *branch;
  int i;

  branch = psp_core_slp_generate_branch_for_request (evt->sip);
  if (branch == NULL)
    {                           /* this message is probably incorrect */
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      return -1;
    }
  req = pspm_slp_find_pending_req (pspm, branch);
  sfree (branch);
  if (req != NULL)
    {
      /* this should be a retransmission of an slp request...       */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "This is a restransmission for a request!\n"));
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);

      if (req->response != NULL && !MSG_IS_STATUS_1XX (req->response))
        {
          /* the remote UA has already answered! so it's more usefull to
             retransmit the response than to transmit again the request!
             This break the rules of a stateless proxy, but anyway, we'll
             increase performance. */
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_WARNING, NULL,
                       "REQUEST received but we already get a final answer for this request! Instead of forwarding it, we send again the answer.\n"));
          i = pspm_slp_forward_response (req->response);
          if (i != 0)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "Cannot forward this response again? Did we really tried before? New answer will be 500!\n"));
              REMOVE_ELEMENT (pspm->slp_pending, req);
              slp_branch_free (req->sl_ctx);
              sfree (req->sl_ctx);
              req->sl_ctx = NULL;
              psp_core_switch_to_uas_mode (req, 500);
              return -1;
            }
          return 0;
        }

      /* check if it is a retransmission?  (in this case, a request exist) */
      if (req->sl_ctx != NULL && req->sl_ctx->request != NULL)
        {
          i = pspm_slp_forward_request (req->sl_ctx->request, NULL);
          /* and error cases? */
          if (i != 0)
            {
              REMOVE_ELEMENT (pspm->slp_pending, req);
              slp_branch_free (req->sl_ctx);
              sfree (req->sl_ctx);
              req->sl_ctx = NULL;
              psp_core_switch_to_uas_mode (req, 500);
              return -1;
            }
          return 0;
        }
      REMOVE_ELEMENT (pspm->slp_pending, req);
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }

  /* cannot find a suitable request in slp. */

  psp_core_event_add_imp_traffic (evt);
  return 0;
}

int
pspm_slp_forward_request (sip_t * request, route_t * route)
{
  int port;
  char *host;

  /* It is possible that a route header have been added... 
     This is only the case if the received request does not
     contains any Route header...  */
  if (route == NULL)
    {
      route = list_get (request->routes, 0);
    }

  if (route == NULL)
    {
      host = request->strtline->rquri->host;
      if (request->strtline->rquri->port == NULL)
        port = 5060;
      else
        port = atoi (request->strtline->rquri->port);
  } else
    {
      host = route->url->host;
      if (route->url->port == NULL)
        port = 5060;
      else
        port = atoi (route->url->port);
    }

  return psp_core_cb_snd_message (NULL, request, host, port, -1);
}

int
pspm_slp_handle_psp_request_4_ack (pspm_slp_t * pspm, psp_req_t * req)
{
  slp_branch_t *slp_branch;
  route_t *route = NULL;
  url_t *rquri = NULL;
  sip_t *fwd;
  int i;

  /* time for handling new requests */
  if (req->locations == NULL || req->locations->url == NULL)
    {
      /* if there is a route header, we use it!
         PLUGINS MUST NOT SEARCH FOR LOCATION, when there is a route header
         in the received request */
      route = list_get (req->request->routes, 0);
      if (route == NULL || route->url == NULL || route->url->host == NULL)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "plugin did not give a valid location!\n"));
          psp_req_free (req);
          sfree (req);
          return -1;
        }
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "Checking if we are responsible for '%s'\n", route->url->host));
      i = psp_core_is_responsible_for_this_route (route->url);
      if (i == 0)               /* yes, it is responsible! So we need to forward it to the next
                                   route element. */
        {
          route = list_get (req->request->routes, 1);
          if (route == NULL)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "plugin did not give a valid location!\n"));
              psp_req_free (req);
              sfree (req);
              return -1;
            }
        }
      /* here, we just have to continue with the normal processing */
    }

  if (route == NULL)
    {
      rquri = req->locations->url;
      /* this is needed to not free the URL when freeing the location context */
      req->locations->url = NULL;
    }

  /* get a copy of the request with the TOP via added */
  i = osip_msg_slp_build_request_to_forward (&fwd, rquri, req->request,
                                             __IS_PSP_STAY_ON_PATH (req->flag));
  if (i != 0)
    {                           /* ??  plugin should be informed that this context failed! */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "slp module: Could not build the ACK request to forward!\n"));
      psp_req_free (req);
      sfree (req);
      return -1;
    }

  i = psp_core_slp_new_branch (&slp_branch, req);
  if (i != 0)
    {                           /* ??  plugin should be informed that this context failed! */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "slp module: Could not build a slp branch for the ACK request to forward!\n"));
      msg_free (fwd);
      sfree (fwd);
      psp_req_free (req);
      sfree (req);
      return -1;
    }
  /* add the branch */
  req->sl_ctx = slp_branch;
  slp_branch->request = fwd;

  i = pspm_slp_call_plugins_for_request (pspm, req);

  i = pspm_slp_forward_request (fwd, route);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "slp module: Could not forward the ACK request!\n"));
      psp_req_free (req);
      sfree (req);
      return -1;
    }
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO1, NULL,
               "slp module: An ACK was forwarded by the slp module!\n"));
  psp_req_free (req);
  sfree (req);
  return 0;

}

int
pspm_slp_handle_psp_request (pspm_slp_t * pspm, psp_req_t * req)
{
  /* processing a stateless request */
  /* step 1: build a basic request to forward
     step 2: call plugins (which complete the request)
     step 3: just send it...
   */
  int i;
  psp_req_t *tmp_req;
  slp_branch_t *slp_branch;
  route_t *route = NULL;
  url_t *rquri = NULL;
  sip_t *fwd;
  char *branch;

  if (req == NULL || req->request == NULL)
    {                           /* impossible... */
      return -1;
    }

  if (MSG_IS_ACK (req->request))
    {                           /* Some INVITE can arrive here! */
      pspm_slp_handle_psp_request_4_ack (pspm, req);
      return 0;
    }

  branch = psp_core_slp_generate_branch_for_request (req->request);
  if (branch == NULL)
    {                           /* this message is probably incorrect */
      psp_req_free (req);
      sfree (req);
      return -1;
    }
  tmp_req = pspm_slp_find_pending_req (pspm, branch);
  sfree (branch);
  if (tmp_req != NULL)
    {
      /* this should be a retransmission of an slp request...       */
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO2, NULL,
                   "This is a restransmission for a request!\n"));
      psp_req_free (req);
      sfree (req);

      if (tmp_req->response != NULL && !MSG_IS_STATUS_1XX (tmp_req->response))
        {
          /* the remote UA has already answered! so it's more usefull to
             retransmit the response than to transmit again the request!
             This break the rules of a stateless proxy, but anyway, we'll
             increase performance. */
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_WARNING, NULL,
                       "REQUEST received but we already get a final answer for this request! Instead of forwarding it, we send again the answer.\n"));
          i = pspm_slp_forward_response (tmp_req->response);
          if (i != 0)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "Cannot forward this response again? Did we really tried before? New answer will be 500!\n"));
              REMOVE_ELEMENT (pspm->slp_pending, tmp_req);
              slp_branch_free (tmp_req->sl_ctx);
              sfree (tmp_req->sl_ctx);
              tmp_req->sl_ctx = NULL;
              psp_core_switch_to_uas_mode (tmp_req, 500);
              return -1;
            }
          return 0;
        }

      if (tmp_req->sl_ctx != NULL && tmp_req->sl_ctx->request != NULL)
        {
          i = pspm_slp_forward_request (tmp_req->sl_ctx->request, NULL);
          /* and error cases? */
          if (i != 0)
            {
              REMOVE_ELEMENT (pspm->slp_pending, tmp_req);
              slp_branch_free (tmp_req->sl_ctx);
              sfree (tmp_req->sl_ctx);
              tmp_req->sl_ctx = NULL;
              psp_core_switch_to_uas_mode (tmp_req, 400);
              return -1;
            }
          return 0;
        }
      REMOVE_ELEMENT (pspm->slp_pending, tmp_req);
      slp_branch_free (tmp_req->sl_ctx);
      sfree (tmp_req->sl_ctx);
      tmp_req->sl_ctx = NULL;
      psp_core_switch_to_uas_mode (tmp_req, 500);
      return -1;
    }

  /* time for handling new requests */
  if (req->locations == NULL || req->locations->url == NULL)
    {
      /* if there is a route header, we use it!
         PLUGINS MUST NOT SEARCH FOR LOCATION, when there is a route header
         in the received request */
      route = list_get (req->request->routes, 0);
      if (route == NULL)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "plugin did not give a valid location!\n"));
          psp_core_switch_to_uas_mode (req, 404);
          return -1;
        }
      if (route == NULL)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "plugin did not give a valid location!\n"));
          psp_req_free (req);
          sfree (req);
          return -1;
        }
      i = psp_core_is_responsible_for_this_route (route->url);
      if (i == 0)               /* yes, it is responsible! So we need to forward it to the next
                                   route element. */
        {
          route = list_get (req->request->routes, 1);
          if (route == NULL)
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "plugin did not give a valid location!\n"));
              psp_req_free (req);
              sfree (req);
              return -1;
            }
        }
      /* here, we just have to continue with the normal processing */
    }

  if (route == NULL)
    {
      rquri = req->locations->url;
      /* this is needed to not free the URL when freeing the location context */
      req->locations->url = NULL;
    }

  /* get a copy of the request with the TOP via added */
  i = osip_msg_slp_build_request_to_forward (&fwd, rquri, req->request,
                                             __IS_PSP_STAY_ON_PATH (req->flag));
  if (i != 0)
    {                           /* ??  plugin should be informed that this context failed! */
      psp_core_switch_to_uas_mode (req, 400);
      return -1;
    }

  i = psp_core_slp_new_branch (&slp_branch, req);
  if (i != 0)
    {                           /* ??  plugin should be informed that this context failed! */
      msg_free (fwd);
      sfree (fwd);
      req->sl_ctx = NULL;
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }
  /* add the branch */
  req->sl_ctx = slp_branch;
  slp_branch->request = fwd;

  /* request to complete is located in req->sl_ctx->request
     it is already cloned by the "imp" module */
  i = pspm_slp_call_plugins_for_request (pspm, req);

  /* The tlp plugins MUST be able to send the reuest even if
     the transaction parameter is NULL...
   */
  i = pspm_slp_forward_request (fwd, route);
  if (i != 0)
    {
      /* MUST SWITCH TO UAS MODE!!!!!!!! */
      slp_branch_free (req->sl_ctx);
      sfree (req->sl_ctx);
      req->sl_ctx = NULL;
      psp_core_switch_to_uas_mode (req, 404);
      return -1;
    }


  /* invite should timeout after 3 mn = 32+148s (instead of default=32s) */
  if (MSG_IS_INVITE (req->request))
    psp_req_add_timeout (req, INVITE_TIMEOUT);
  /* we'll reset it to 5s if the call is answerd succesfully to give
     enough time to manage retransmissions */
  else if (MSG_IS_ACK (req->request))
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_INFO1, NULL,
                   "slp module: An ACK was forwarded by the slp module!\n"));
      /* we don't keep ACK context */
      psp_req_free (req);
      sfree (req);
      return 0;
    }

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO2, NULL,
               "slp module: Adding element in slp_pending!\n"));

  /* time to store the context.. */
  ADD_ELEMENT (pspm->slp_pending, req);
  return 0;
}

int
pspm_slp_handle_psp_response (pspm_slp_t * pspm, sip_t * response)
{
  via_t *via;
  generic_param_t *b;
  int i;
  psp_req_t *req;
  psp_req_t *element;

  /* find the corresponding psp_req_t element */

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO4, NULL,
               "slp module: entering pspm_slp_handle_psp_response!\n"));

  via = list_get (response->vias, 0);
  i = via_param_getbyname (via, "branch", &b);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "slp module: Response has a bad via feild!\n"));
      msg_free (response);
      sfree (response);
      return -1;
    }
  req = NULL;
  for (element = pspm->slp_pending; element != NULL; element = element->next)
    {
      if (0 == slp_branch_match (element->sl_ctx, b->gvalue))
        {
          req = element;
          break;
        }
    }

  if (req == NULL)
    {
      sip_t *answer;

      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_WARNING, NULL,
                   "No psp_req for response in slp!\n"));
      /* we should forward it anyway... */
      i = osip_msg_slp_build_response_to_forward (&answer, response);
      if (i != 0)
        {
          msg_free (response);
          sfree (response);
          return -1;
        }
      i = pspm_slp_forward_response (answer);
      msg_free (response);
      sfree (response);
      msg_free (answer);
      sfree (answer);
      if (i != 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "Could not forward response!\n"));
          return -1;
        }
      return 0;
    }

  /* verify it's not a retransmission... (to avoid calling plugins again) */
  if (req->response != NULL)
    {
      /*    continue !   */
      if (0 == strcmp (req->response->strtline->statuscode,
                       response->strtline->statuscode))
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_INFO2, NULL,
                       "Module slp detects a response retransmission!\n"));
          /* so we don't have to build it again */
          msg_free (response);
          sfree (response);
          i = pspm_slp_forward_response (req->response);
          if (i != 0)
            {
              REMOVE_ELEMENT (pspm->slp_pending, req);  /* PROTOECT WITH MUTEX! */
              slp_branch_free (req->sl_ctx);
              sfree (req->sl_ctx);
              req->sl_ctx = NULL;
              req->response = NULL;
              psp_core_switch_to_uas_mode (req, 500);
              return -1;
            }
          return 0;             /* not an error case */
      } else if (0 > strcmp (req->response->strtline->statuscode,
                             response->strtline->statuscode))
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO1, NULL,
                                  "slp module: new response status for transaction old:%s new: %s!\n",
                                  req->response->strtline->statuscode,
                                  response->strtline->statuscode));
          /* continue processing */
          msg_free (req->response);
          sfree (req->response);
          req->response = NULL;
      } else
        {
          OSIP_TRACE (osip_trace (__FILE__, __LINE__, OSIP_INFO2, NULL,
                                  "slp module: new response but useless to forward old:%s new: %s!\n",
                                  req->response->strtline->statuscode,
                                  response->strtline->statuscode));
          /* just discard the response as it is too old for this transaction */
          msg_free (response);
          sfree (response);
          return 0;             /* not an error case */
        }
    }

  /* now, req->response is NULL */
  i = osip_msg_slp_build_response_to_forward (&(req->response), response);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Could not build response to forward in slp!\n"));
      /* MUST SWITCH TO UAS MODE!!!!!!!! */
      REMOVE_ELEMENT (pspm->slp_pending, req);
      slp_branch_free (req->sl_ctx);
      sfree (req->sl_ctx);
      req->sl_ctx = NULL;
      req->response = NULL;
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }

  req->sl_ctx->response = response;
  i = pspm_slp_call_plugins_for_rcv_response (pspm, req);
  req->sl_ctx->response = NULL;

  /* in all cases, we don't need the response anymore */
  msg_free (response);
  sfree (response);
  response = NULL;

  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Failed to call plugin for rcv response slp!\n"));
      REMOVE_ELEMENT (pspm->slp_pending, req);
      slp_branch_free (req->sl_ctx);
      sfree (req->sl_ctx);
      req->sl_ctx = NULL;
      req->response = NULL;
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }
  req->sl_ctx->response = response;
  i = pspm_slp_call_plugins_for_snd_response (pspm, req);
  req->sl_ctx->response = NULL;
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Failed to call plugin for snd response slp!\n"));
      /* MUST SWITCH TO UAS MODE!!!!!!!! */
      REMOVE_ELEMENT (pspm->slp_pending, req);
      slp_branch_free (req->sl_ctx);
      sfree (req->sl_ctx);
      req->sl_ctx = NULL;
      req->response = NULL;
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }

  i = pspm_slp_forward_response (req->response);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "slp module: could not send response!\n"));
      REMOVE_ELEMENT (pspm->slp_pending, req);
      slp_branch_free (req->sl_ctx);
      sfree (req->sl_ctx);
      req->sl_ctx = NULL;
      req->response = NULL;
      psp_core_switch_to_uas_mode (req, 500);
      return -1;
    }
  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO3, NULL,
               "slp module: response forwarded!\n"));

  if (!MSG_IS_STATUS_1XX (req->response))
    /* we set the timeout to be 20 seconds from now.
       It should be enough to handle any retransmissions. */
    psp_req_set_timeout (req, 20);

  return 0;
}

int
pspm_slp_call_plugins_for_request (pspm_slp_t * pspm, psp_req_t * req)
{
  slp_fwd_func_tab_t *t;
  slp_fwd_func_t *f;

  if (MSG_IS_INVITE (req->sl_ctx->request))
    t = pspm->fwd_invites;
  else if (MSG_IS_ACK (req->sl_ctx->request))
    t = pspm->fwd_acks;
  else if (MSG_IS_REGISTER (req->sl_ctx->request))
    t = pspm->fwd_registers;
  else if (MSG_IS_BYE (req->sl_ctx->request))
    t = pspm->fwd_byes;
  else if (MSG_IS_OPTIONS (req->sl_ctx->request))
    t = pspm->fwd_optionss;
  else if (MSG_IS_INFO (req->sl_ctx->request))
    t = pspm->fwd_infos;
  else if (MSG_IS_CANCEL (req->sl_ctx->request))
    t = pspm->fwd_cancels;
  else if (MSG_IS_NOTIFY (req->sl_ctx->request))
    t = pspm->fwd_notifys;
  else if (MSG_IS_SUBSCRIBE (req->sl_ctx->request))
    t = pspm->fwd_subscribes;
  else
    t = pspm->fwd_unknowns;

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_fwd_request_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}

int
pspm_slp_call_plugins_for_rcv_response (pspm_slp_t * pspm, psp_req_t * req)
{
  slp_rcv_func_tab_t *t;
  slp_rcv_func_t *f;

  if (MSG_IS_STATUS_1XX (req->sl_ctx->response))
    t = pspm->rcv_1xxs;
  else if (MSG_IS_STATUS_2XX (req->sl_ctx->response))
    t = pspm->rcv_2xxs;
  else if (MSG_IS_STATUS_3XX (req->sl_ctx->response))
    t = pspm->rcv_3xxs;
  else if (MSG_IS_STATUS_4XX (req->sl_ctx->response))
    t = pspm->rcv_4xxs;
  else if (MSG_IS_STATUS_5XX (req->sl_ctx->response))
    t = pspm->rcv_5xxs;
  else if (MSG_IS_STATUS_6XX (req->sl_ctx->response))
    t = pspm->rcv_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_rcv_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}


int
pspm_slp_call_plugins_for_snd_response (pspm_slp_t * pspm, psp_req_t * req)
{
  slp_snd_func_tab_t *t;
  slp_snd_func_t *f;

  if (MSG_IS_STATUS_1XX (req->response))
    t = pspm->snd_1xxs;
  else if (MSG_IS_STATUS_2XX (req->response))
    t = pspm->snd_2xxs;
  else if (MSG_IS_STATUS_3XX (req->response))
    t = pspm->snd_3xxs;
  else if (MSG_IS_STATUS_4XX (req->response))
    t = pspm->snd_4xxs;
  else if (MSG_IS_STATUS_5XX (req->response))
    t = pspm->snd_5xxs;
  else if (MSG_IS_STATUS_6XX (req->response))
    t = pspm->snd_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_snd_answer_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}
