/*
  The partysip program is a modular SIP proxy server (SIP -rfc2543-)
  Copyright (C) 2002  WellX Telecom   - <partysip@wellx.com>
  Copyright (C) 2002  Aymeric MOIZARD - <jack@atosc.org>
  
  Partysip is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  Partysip is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with Foobar; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <partysip.h>
#include <osip_ctx.h>
#include <osip_msg.h>

int
pspm_uap_init (pspm_uap_t ** pspm)
{
  int i;

  (*pspm) = (pspm_uap_t *) smalloc (sizeof (pspm_uap_t));
  if (*pspm == NULL)
    return -1;

  i = module_init (&((*pspm)->module), MOD_THREAD, "uap");
  if (i != 0)
    goto mui_error1;

  (*pspm)->uap_plugins = NULL;  /* Added by Tim Na */

  /*  (*pspm)->uap_pending = NULL; */

  (*pspm)->uas_request = (fifo_t *) smalloc (sizeof (fifo_t));
  if ((*pspm)->uas_request == NULL)
    goto mui_error2;

  fifo_init ((*pspm)->uas_request);

  (*pspm)->func_1xxs = NULL;
  (*pspm)->func_2xxs = NULL;
  (*pspm)->func_3xxs = NULL;
  (*pspm)->func_4xxs = NULL;
  (*pspm)->func_5xxs = NULL;
  (*pspm)->func_6xxs = NULL;

  i = uap_func_tab_init (&((*pspm)->func_1xxs));
  if (i != 0)
    goto mui_error3;
  i = uap_func_tab_init (&((*pspm)->func_2xxs));
  if (i != 0)
    goto mui_error3;
  i = uap_func_tab_init (&((*pspm)->func_3xxs));
  if (i != 0)
    goto mui_error3;
  i = uap_func_tab_init (&((*pspm)->func_4xxs));
  if (i != 0)
    goto mui_error3;
  i = uap_func_tab_init (&((*pspm)->func_5xxs));
  if (i != 0)
    goto mui_error3;
  i = uap_func_tab_init (&((*pspm)->func_6xxs));
  if (i != 0)
    goto mui_error3;

  return 0;

mui_error3:
  module_release ((*pspm)->module);
  pspm_uap_free (*pspm);
  sfree (*pspm);
  return -1;
mui_error2:
  module_release ((*pspm)->module);
  module_free ((*pspm)->module);
  sfree ((*pspm)->module);
mui_error1:
  sfree (*pspm);
  return -1;
}

void
pspm_uap_free (pspm_uap_t * pspm)
{
  uap_plugin_t *p;
  psp_req_t *req;

  if (pspm == NULL)
    return;

  uap_func_tab_free (pspm->func_1xxs);
  uap_func_tab_free (pspm->func_2xxs);
  uap_func_tab_free (pspm->func_3xxs);
  uap_func_tab_free (pspm->func_4xxs);
  uap_func_tab_free (pspm->func_5xxs);
  uap_func_tab_free (pspm->func_6xxs);

  sfree (pspm->func_1xxs);
  sfree (pspm->func_2xxs);
  sfree (pspm->func_3xxs);
  sfree (pspm->func_4xxs);
  sfree (pspm->func_5xxs);
  sfree (pspm->func_6xxs);

  module_free (pspm->module);
  sfree (pspm->module);

  req = fifo_tryget (pspm->uas_request);
  while (req != NULL)
    {
      psp_req_free (req);
      sfree (req);
      req = fifo_tryget (pspm->uas_request);
    }

  fifo_free (pspm->uas_request);
  sfree (pspm->uas_request);

  /*
     for (req=pspm->uap_pending;req!=NULL;req=pspm->uap_pending)
     {
     REMOVE_ELEMENT(pspm->uap_pending, req);
     psp_req_free(req);
     sfree(req);
     }
   */
  for (p = pspm->uap_plugins; p != NULL; p = pspm->uap_plugins)
    {
      REMOVE_ELEMENT (pspm->uap_plugins, p->next);
      uap_plugin_free (p);
      sfree (p);
    }

}

int
pspm_uap_release (pspm_uap_t * pspm)
{
  int i;

  i = module_release (pspm->module);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_uap_start (pspm_uap_t * pspm, void *(*func_start) (void *), void *arg)
{
  int i;

  i = module_start (pspm->module, func_start, arg);
  if (i != 0)
    return -1;
  return 0;
}

int
pspm_uap_execute (pspm_uap_t * pspm, int sec_max, int usec_max, int max)
{
  psp_req_t *req;
  fd_set uap_fdset;
  int max_fd;
  struct timeval tv;

  if (pspm == NULL)
    return -1;

  while (1)
    {
      int s;
      int i;

      tv.tv_sec = sec_max;
      tv.tv_usec = usec_max;

      max--;
      FD_ZERO (&uap_fdset);
      s = ppl_pipe_get_read_descr (pspm->module->wakeup);
      if (s <= 1)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "No wakeup value in module uap!\n"));
          return -1;
        }
      FD_SET (s, &uap_fdset);

      max_fd = s;

      if ((sec_max == -1) || (usec_max == -1))
        i = select (max_fd + 1, &uap_fdset, NULL, NULL, NULL);
      else
        i = select (max_fd + 1, &uap_fdset, NULL, NULL, &tv);

      if (FD_ISSET (s, &uap_fdset))
        {                       /* time to wake up: we receive "q" to quit and any other char to read all socket */
          char tmp[2];

          i = ppl_pipe_read (pspm->module->wakeup, tmp, 1);
          if (i == 1 && tmp[0] == 'q')
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_INFO2, NULL,
                           "Exiting uap module!\n"));
              return 0;
            }
        }

      req = fifo_tryget (pspm->uas_request);
      while (req != NULL)
        {
          pspm_uap_handle_psp_request (pspm, req);
          req = fifo_tryget (pspm->uas_request);
        }

      if (max == 0)
        return 0;

    }
}


int
pspm_uap_handle_psp_request (pspm_uap_t * pspm, psp_req_t * req)
{
  /* "req" is sent by imp and is a copy of the NEW REQUEST */
  /* step 1: build a basic response
     step 2: call plugins (which complete the response)
     step 3: give it to osip
   */
  transaction_t *transaction;
  sip_t *resp;
  sipevent_t *evt;
  osip_ctx_t *osip_ctx;
  sipevent_t *evt_request;
  int i;
  int xxx;

  i = osip_msg_uap_build_response (&resp, req->uas_status, req->request);
  if (i != 0)
    {
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "uap module: Could not build response for uap transaction!\n"));
      psp_req_free (req);
      sfree (req);
      return -1;
    }

  /* if the psp_req_t element contains locations and is a 3xx, insert locations */
  xxx = psp_req_get_uas_status (req);
  if (300 <= xxx && xxx <= 399)
    {
      location_t *loc;

      for (loc = req->locations; loc != NULL; loc = loc->next)
        {
          char *tmp;            /* convert to contact header */

          if (loc->url != NULL && 0 == url_2char (loc->url, &tmp))      /* else error??? */
            {
              i = msg_setcontact (resp, tmp);   /* insert contact in SIP message */
              if (i != 0)
                {
                  OSIP_TRACE (osip_trace
                              (__FILE__, __LINE__, OSIP_WARNING, NULL,
                               "Bad contat header given by plugin to module uap!\n"));
                }
          } else
            {
              OSIP_TRACE (osip_trace
                          (__FILE__, __LINE__, OSIP_ERROR, NULL,
                           "No contact header given by plugin to module uap (but 3xx answer asked?)!\n"));
            }
        }
      if (list_size (resp->contacts) == 0)
        {
          OSIP_TRACE (osip_trace
                      (__FILE__, __LINE__, OSIP_ERROR, NULL,
                       "A plugin ask for redirection without giving a contact!\n"));
          /* so we change the status code to 404 Not found */
          psp_req_set_uas_status (req, 404);
          sfree (resp->strtline->statuscode);
          msg_setstatuscode (resp, sgetcopy ("404"));
          sfree (resp->strtline->reasonphrase);
          msg_setreasonphrase (resp, msg_getreason (404));
        }
    }

  req->response = resp;
  pspm_uap_call_plugins (pspm, req);


  /* send a copy of the response to the osip stack */
  evt = osip_new_outgoing_sipmessage (resp);
  req->response = NULL;
  evt->transactionid = req->osip_id;

  i = psp_core_find_transaction_and_add_event (evt);
  if (i == 0)
    {
      psp_req_free (req);
      sfree (req);
      return 0;
    }

  OSIP_TRACE (osip_trace
              (__FILE__, __LINE__, OSIP_INFO4, NULL,
               "We have to start a transaction (may be this request comes from slp or sfp module!!!\n"));
  evt_request = osip_new_incoming_sipmessage (req->request);
  req->request = NULL;          /* so it's no free twice */
  evt_request->transactionid = 0;
  transaction = psp_core_uap_create_transaction (evt_request);
  if (transaction == NULL)
    {
      msg_free (evt->sip);
      sfree (evt->sip);
      sfree (evt);
      msg_free (evt_request->sip);
      sfree (evt_request->sip);
      sfree (evt_request);
      psp_req_free (req);
      sfree (req);
      OSIP_TRACE (osip_trace
                  (__FILE__, __LINE__, OSIP_ERROR, NULL,
                   "Could not create transaction!!!\n"));
      return -1;
    }
  evt_request->transactionid = transaction->transactionid;
  fifo_add (transaction->transactionff, evt_request);

  psp_req_set_osip_id (req, transaction->transactionid);

  /* moved above osip_ctx_start */
  evt->transactionid = transaction->transactionid;
  fifo_add (transaction->transactionff, evt);

  osip_ctx_init (&osip_ctx, transaction);
  osip_ctx_start (osip_ctx);


  psp_req_free (req);
  sfree (req);
  return 0;
}

int
pspm_uap_call_plugins (pspm_uap_t * pspm, psp_req_t * req)
{
  uap_func_tab_t *t;
  uap_func_t *f;

  if (MSG_IS_STATUS_1XX (req->response))
    t = pspm->func_1xxs;
  else if (MSG_IS_STATUS_2XX (req->response))
    t = pspm->func_2xxs;
  else if (MSG_IS_STATUS_3XX (req->response))
    t = pspm->func_3xxs;
  else if (MSG_IS_STATUS_4XX (req->response))
    t = pspm->func_4xxs;
  else if (MSG_IS_STATUS_5XX (req->response))
    t = pspm->func_5xxs;
  else if (MSG_IS_STATUS_6XX (req->response))
    t = pspm->func_6xxs;
  else
    {
      printf ("ERROR: I do not understand this response");
      return -1;
    }

  for (f = t->func_hook_really_first; f != NULL; f = f->next)
    {
      f->cb_snd_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_first; f != NULL; f = f->next)
    {
      f->cb_snd_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_middle; f != NULL; f = f->next)
    {
      f->cb_snd_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_last; f != NULL; f = f->next)
    {
      f->cb_snd_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  for (f = t->func_hook_really_last; f != NULL; f = f->next)
    {
      f->cb_snd_func (req);
      if (__IS_PSP_STOP (req->flag))
        return 0;
    }
  return 0;
}
